import os
import numpy as np
from shutil import rmtree
from .hclust import clusterTools as CT
from tempfile import mkdtemp
from pathlib import Path
from ..utils import read_bundle, write_bundle, sampling
from .hclust.c_wrappers import fiberDistanceMax, getAffinityGraphFromDistanceMatrix, getAverageLinkHCFromGraphFile
import logging
import time
import joblib
from .ffclust.clustering import split_fibers, parallel_points_clustering, MapClustering,\
    small_clusters_reassignment, get_groups, parallel_group_join_clique
from .ffclust.utils import clusters_fibers, save_clusters, save_clusters_centroids

"""
HClust Sub-Module.
"""

def hierarchical(fiber_input, outfile_dir, MaxDistance_Thr, PartDistance_Thr, variance, clusters_id_dir):
    """
    Run to Hierarchical clustering
    """
    matrix_distance_dir = os.path.join(outfile_dir, 'distance_matrix.bin')
    affinity_graph_dir = os.path.join(outfile_dir,'affinity_graph.txt')
    dendogram_dir = os.path.join(outfile_dir,'dendogram.txt')

    # Distance matrix
    fiberDistanceMax(fiber_input, matrix_distance_dir)

    # affinity graph
    getAffinityGraphFromDistanceMatrix(matrix_distance_dir, affinity_graph_dir, MaxDistance_Thr)

    # Dendogram
    getAverageLinkHCFromGraphFile(affinity_graph_dir, dendogram_dir)

    # Dendogram partition
    wfv=CT.wforest_partition_maxdist_from_graph(dendogram_dir,PartDistance_Thr,True,affinity_graph_dir,variance)
    clusters=wfv.clusters

    ar=open(clusters_id_dir,'wt')
    ar.write(str(clusters))
    ar.close()

    return clusters



def write_fiber_clusters (fiber_input, clusters, final_bundles_dir):

    """Writes the fiber clusters
    """  
    os.makedirs(final_bundles_dir, exist_ok=True)
    
    for i, j in enumerate(clusters):
        write_bundle(os.path.join(final_bundles_dir, f"{i}.bundles"), fiber_input[j])



def is_reversed(cluster):
    base = np.stack([cluster, cluster[:, ::-1]])
    x = base - cluster[None, 0]
    x = np.square(x)
    x = np.sum(x, axis=-1)
    x = np.max(x, axis=-1)
    return x[0] > x[1]

def cal_centroide(cluster):
    cluster = np.asarray(cluster)
    r = is_reversed(cluster)
    if np.any(r):
        c_copy = np.empty_like(cluster)
        c_copy[~r] = cluster[~r]
        c_copy[r] = cluster[r, ::-1]
        cluster = c_copy
    return np.sum(cluster, axis=0) / len(cluster)

def write_centroids(clusters_dir, file_out):
    p = Path(clusters_dir)
    centroids = []
    clusters_paths = sorted(p.glob("*.bundles"), key=lambda x: int(x.stem))
    for cluster_path in clusters_paths:
        cluster = read_bundle(str(cluster_path))
        centroids.append(cal_centroide(cluster))
    write_bundle(file_out, centroids)


def hclust(file_in: str, dir_out: str, fiber_thr: int, partition_thr: int, variance: int) -> None:
    """
    Average-link hierarchical agglomerative clustering algorithm which allows finding bundles based on a pairwise fiber distance measure.

    Parameters
    ----------
    file_in : str
        Tractography dataset file in *bundles* format.
    dir_out : str
        Directory to store all the results generated by the algorithm.
    fiber_thr : str
        Maximum distance threshold in *mm*, default: *70*.
    partition_thr : str
        Partition threshold in *mm*, default: *70*.
    variance : str
        Variance squared and provides a similarity scale in *mm*, default: *3600*.

    Returns
    -------
    None

    Notes
    -----
    This function generates the following files in the specified directory:

    final_bundles : bundles files
        Directory that stores all the fiber clusters identified in separated datasets (*bundles/bundlesdata* files), sampled at 21 points. 
        The file names are labeled with integer numbers ranging from zero to the total number of fiber clusters (N-1) identified.
    final_bundles_allpoints : bundles files    
        Directory that stores all the fiber clusters identified in separated datasets (*bundles/bundlesdata* files). 
        The file names are labeled with integer numbers ranging from zero to the total number of fiber clusters (N-1) identified.
        This directory is only generated when the input tractography dataset file has not been sampled at 21 points. 
    centroids : bundles files
        A directory storing two files corresponding to a tractography dataset in *bundles/bundlesdata* format, 
        containing one centroid for each created cluster. This dataset is named *centroids.bundles/centroids.bundlesdata*, and is sampled at 21 points in the atlas space (MNI).
    bundles_id : text file
        Text file that stores the fiber indexes from the original input tractography dataset file for each detected cluster. 
        Each line in the file correspond to a cluster in correlative order.
    outputs : Several formats
        Temporal directory with intermediate results, such as the distance matrix, the affinity graph, and the dendrogram.

    """
    
    if os.path.exists(dir_out):
        rmtree(dir_out)

    os.mkdir(dir_out)

    final_bundles21p_dir = os.path.join(dir_out, 'final_bundles')
    os.makedirs(final_bundles21p_dir, exist_ok=True)

    final_centroids_dir = os.path.join(dir_out, 'centroids')
    final_centroids_file = os.path.join(final_centroids_dir, 'centroids.bundles')
    os.makedirs(final_centroids_dir, exist_ok=True)

    outfile_dir = os.path.join(dir_out, 'outputs')
    os.makedirs(outfile_dir, exist_ok=True)

    clusters_id_dir = os.path.join(dir_out, 'bundles_id.txt')

    num_point_fibers = 1
    data = read_bundle(file_in)
    for fib in data:
        if len(fib) != 21:
            num_point_fibers = 0
            break

    if num_point_fibers == 0:

        fibers21p = os.path.join(outfile_dir, 'fibers_21p.bundles')
        sampling(file_in, fibers21p, 21)
       
        clusters = hierarchical(fibers21p, outfile_dir, fiber_thr, partition_thr, variance, clusters_id_dir)

        final_bundles_allpoints_dir = os.path.join(dir_out, 'final_bundles_allpoints')
        os.makedirs(final_bundles_allpoints_dir, exist_ok=True)

        fibers_allpoints = np.asarray(data, dtype = object)
        write_fiber_clusters (fibers_allpoints, clusters, final_bundles_allpoints_dir)

        fibers_21points = np.asarray(read_bundle(fibers21p))
        write_fiber_clusters (fibers_21points, clusters, final_bundles21p_dir)

        write_centroids(final_bundles21p_dir, final_centroids_file)

    else:

        clusters = hierarchical(file_in, outfile_dir, fiber_thr, partition_thr, variance, clusters_id_dir)   
        
        fibers_input = np.asarray(data)     
        write_fiber_clusters (fibers_input, clusters, final_bundles21p_dir)        
        
        write_centroids(final_bundles21p_dir, final_centroids_file)



def ffclust(file_in: str, dir_out: str, points: [0,3,10,17,20], ks: [200,300,300,300,200], assign_thr: int=6, join_thr: int=6) -> bool:
    """
    Intra-subject clustering algorithm aims to identify compact and homogeneous fiber clusters on a large tractography dataset.

    Parameters
    ----------
    file_in : str
        Tractography data file in *'.bundles'* format.
    dir_out : str
        Directory to store all the results generated by the algorithm.
    points : list()
        Index of the points to be used in the point clustering (K-means), default: 0, 3, 10, 17, 20.
    ks : list()
        Number of clusters to be computed for each point using K-Means, default: 300, 200, 200, 200, 300.
    assign_thr : str
        Maximum distance threshold for the cluster reassignment in *mm*, default: *6*.
    join_thr : str
        Maximum distance threshold for the cluster merge in *mm*, default: *6*.

    Returns
    -------
    None

    Notes
    -----
    This function generates the following files in the specified directory:

    final_bundles : bundles files
        Directory that stores all the fiber clusters identified in separated datasets (*bundles/bundlesdata* files), sampled at 21 points. 
        The file names are labeled with integer numbers ranging from zero to the total number of fiber clusters (N-1) identified.
    final_bundles_allpoints : bundles files    
        Directory that stores all the fiber clusters identified in separated datasets (*bundles/bundlesdata* files). 
        The file names are labeled with integer numbers ranging from zero to the total number of fiber clusters (N-1) identified.
        This directory is only generated when the input tractography dataset file has not been sampled at 21 points. 
    centroids : bundles files
        A directory storing two files corresponding to a tractography dataset in *bundles/bundlesdata* format, 
        containing one centroid for each created cluster. This dataset is named *centroids.bundles/centroids.bundlesdata*, and is sampled at 21 points in the atlas space (MNI).
    bundles_id : text file
        Text file that stores the fiber indexes from the original input tractography dataset file for each detected cluster. 
        Each line in the file correspond to a cluster in correlative order.
    outputs : Several formats
        Temporal directory with intermediate results, such as the point clusters.       
    """

    if isinstance(assign_thr, str):
        assign_thr = int(assign_thr)
    if isinstance(join_thr, str):
        join_thr = int(join_thr)

    point = np.array(points)
    kss = np.array(ks)

    os.makedirs(dir_out, exist_ok=True)

    final_bundles21p_dir = os.path.join(dir_out, 'final_bundles')
    os.makedirs(final_bundles21p_dir, exist_ok=True)

    final_centroids_dir = os.path.join(dir_out,'centroids')
    os.makedirs(final_centroids_dir, exist_ok=True)

    object_dir = os.path.join(dir_out, 'outputs')
    os.makedirs(object_dir, exist_ok=True)
    data=read_bundle(file_in)

    num_point_fibers=1
    for fib in data:
        if len(fib) != 21:
            num_point_fibers=0
            break


    if num_point_fibers == 0:
        fibers21p = os.path.join(object_dir,'fiber_input_21pts.bundles')
        sampling(file_in, fibers21p, 21)
        fibers = np.asarray(read_bundle(fibers21p))
    
    else:
        fibers = np.asarray(data)

    # New logging file each time, changing filemode to a should append instead.
    logging.basicConfig(filename=os.path.join(object_dir,'info.log'), filemode='w', level=logging.INFO)

    t1 = time.time()
    X = split_fibers(fibers[:,point,:],point)
    labels, clusterers = parallel_points_clustering(X=X, ks=kss)
    logging.info('Kmeans Time: {}'.format(time.time() - t1))

    t1 = time.time()
    m = MapClustering()
    map_clusters = m.cluster(fibers, labels)
    logging.info('Map Time: {}'.format(time.time() - t1))

    map_output_filename = os.path.join(object_dir, 'clusters_map.txt')
    save_clusters(dataset=fibers, clusters=map_clusters, filename=map_output_filename)

    with open(os.path.join(object_dir,'stats.txt'), 'w') as f:
        f.write('Number of clusters in map_clusters: {}\n'.format(len(map_clusters)))
        f.write('Number of fibers in map_clusters: {}\n'.format(sum(map_clusters.clusters_sizes())))

    for p,k, clusterer in zip(point, kss, clusterers):
        joblib.dump(clusterer, os.path.join (object_dir, 'clusterer-Point{}-k{}.pkl'.format(p,k)))

  
    t1 = time.time()
    actual_clusters = small_clusters_reassignment(clusters=map_clusters,
                                                  min_size_filter=6,
                                                  max_size_filter=5,
                                                  input_dir = 'segmentation/bundles/result/parallelFastCPU',
                                                  threshold = assign_thr,
                                                  refdata = fibers)

    ident_clusters = dict()
    for i,c in enumerate(actual_clusters):
        ident_clusters[str(c.indices)] = i

    logging.info('Segmentation Time {}'.format(time.time() - t1))

    t1 = time.time()
    point_index = 10
    ngroups = kss[len(kss)//2]

    centroids = np.asarray([x.centroid for x in actual_clusters])
    centroids_points = centroids[:,point_index]
    clusterer=clusterers[len(point)//2]
    labels = clusterer.predict(centroids_points)

    groups = get_groups(labels, ngroups=ngroups)

    final_clusters = parallel_group_join_clique(actual_clusters, groups, fibers, final_bundles21p_dir, dir_out, final_centroids_dir, ident_clusters, object_dir, join_thr)
    final_centroids_filename = os.path.join(object_dir, 'centroids.txt')
    save_clusters_centroids(clusters=final_clusters, filename=final_centroids_filename)

    os.makedirs(final_bundles21p_dir, exist_ok=True)
    clusters_fibers (final_clusters, final_bundles21p_dir, fibers)

    final_bundles_allpoints_dir = os.path.join(dir_out, 'final_bundles_allpoints')
    os.makedirs(final_bundles_allpoints_dir, exist_ok=True)

    if num_point_fibers == 0:
        clusters_fibers (final_clusters, final_bundles_allpoints_dir, np.asarray(data, dtype=object))
    

    return True



"""Segmentation Module
"""

import os
import sys
import numpy as np
from pathlib import Path
import shutil
from .fiberseg.c_wrappers import segment
from ..utils import sampling, read_bundle, write_bundle

def is_reversed(cluster):
    base = np.stack([cluster, cluster[:, ::-1]])
    x = base - cluster[None, 0]
    x = np.square(x)
    x = np.sum(x, axis=-1)
    x = np.max(x, axis=-1)
    return x[0] > x[1]

def cal_centroide(cluster):
    cluster = np.asarray(cluster)
    r = is_reversed(cluster)
    if np.any(r):
        c_copy = np.empty_like(cluster)
        c_copy[~r] = cluster[~r]
        c_copy[r] = cluster[r, ::-1]
        cluster = c_copy
    return np.sum(cluster, axis=0) / len(cluster)

def write_centroids(clusters_dir, file_out):
    p = Path(clusters_dir)
    centroids = []
    clusters_paths = p.glob("*.bundles")
    for cluster_path in clusters_paths:
        cluster = read_bundle(str(cluster_path)) 
        c = cal_centroide(cluster)
        centroids.append(c.astype(np.float32))
        
    write_bundle(file_out, centroids)


def fiberseg(file_in: str, subj_name: str, atlas_dir: str, atlas_info: str, dir_out: str) -> None:
    """
    White matter fiber bundle segmentation algorithm based on a multi-subject atlas.

    Parameters
    ----------
    file_in : str
        Tractography dataset file in *bundles* format.
    subj_name : str
        Subject name, used to label the results.
    atlas_dir : str
        Bundle atlas directory, with bundles in separated files, sampled at 21 equidistant points. The bundle atlases provided are in same folders.
    atlas_info : str
        Text file associated to the used atlas, that stores information needed to apply the segmentation algorithm, i.e., a list of the atlas fascicles, containing the name, the segmentation
        threshold (in *mm*) and the size of each fascicle. Note that the segmentation threshold can be adjusted depending on the database to be used.
    dir_out : str
        Directory name to store all the results generated by the algorithm.

    Return
    ------
    None

    Notes
    -----
    This function generates the following files in the specified directory:

    final_bundles : bundles files
        The directory contains individual files for each segmented fascicle in the *bundles/bundlesdata* format, sampled at 21 points or all point, and in the atlas space (MNI). 
        Each file is named using the subject's name followed by the atlas label. 
    centroids : bundles files
        A directory storing two files corresponding to a tractography dataset in *bundles/bundlesdata* format, containing one centroid for each segmented fascicle. 
        This dataset is named *centroids.bundles/centroids.bundlesdata*, and is sampled at 21 points in the atlas space (MNI).
    bundles_id : text files
        The directoty that has a text file for each segmented bundle. A text file of a segmented bundle contains the indices of the fibers in the original  subject's tractography dataset file. 
        The name of each file corresponds to the name of the segmented bundle file with ``'.txt'`` extension.
        These files are used to obtain the segmented fascicles in the subject's original space and with all points of brain fibers.

    """

    id_seg_result= os.path.join(dir_out,'bundles_id')
    os.makedirs(id_seg_result, exist_ok=True)

    has21points = True
    data=read_bundle(file_in)

    for fib in data:
        if len(fib) != 21:
            has21points = False
            break

    final_bundles21p_dir = os.path.join(dir_out, 'final_bundles')
    os.makedirs(final_bundles21p_dir, exist_ok=True)

    centroid_dir= os.path.join(dir_out,'centroids')
    os.makedirs(centroid_dir, exist_ok=True)    

    final_bundles_dir = os.path.join(dir_out, 'final_bundles')
    os.makedirs(final_bundles_dir, exist_ok=True)    

    if not has21points:

        outfile_dir= os.path.join(dir_out, 'outputs')
        os.makedirs(outfile_dir, exist_ok=True)

        fibers21p = os.path.join(outfile_dir,'fibers_21p.bundles')

        sampling(file_in, fibers21p, 21)

        segment(21, fibers21p, subj_name, atlas_dir, atlas_info,
                final_bundles21p_dir, id_seg_result)

        shutil.rmtree(outfile_dir)
               
    else:
        segment(21, file_in, subj_name, atlas_dir, atlas_info,
                final_bundles21p_dir, id_seg_result)  


    write_centroids(os.path.join(dir_out, 'final_bundles'), os.path.join(centroid_dir,'centroids.bundles'))


import numpy as np
from .SegmentationHandler import SegmentationHandler
from ..BoundingBox import BoundingBox
from ..Tools.visualizationEnums import SegmentationTypes
from ...FiberVis_core import inPlaceSegmentationMethod

class InPlaceSegmentation(SegmentationHandler):
    def __init__(self, bundle, shaderDict):
        super().__init__(bundle, shaderDict)

        self.segmentationIdentifier = SegmentationTypes.InPlace

        self.bundleStates = np.ones(len(self.bundlesName), dtype=np.int32) # int8 = c boolean
        self.percentage = 100
        self.fileName = 'In place segmentation' # temporal

        self.alpha = 0.8

        # Creates VBOs & an EBO, then populates them with the point, normal and color data. The elements data goes into the EBO
        self.configFiberValidator()

        self._loadBuffers()
        self.buildVertex2Fiber()
        self.vboAndLinkVertex2Fiber()

        self.boundingbox = BoundingBox(shaderDict, self, bundle.boundingbox.dims, bundle.boundingbox.center)


    def segmentMethod(self):
        self.fiberValidator[:self.curvescount] = 0

        inPlaceSegmentationMethod(
                self.bundleStates.size,
                self.percentage,
                self.bundlesStart,
                self.bundleStates,
                self.fiberValidator)

        # cfuncs.inPlaceSegmentationMethod(
        #       self.bundleStates.size,
        #       self.percentage,
        #       self.bundlesStart.ctypes.data,
        #       self.bundleStates.ctypes.data,
        #       self.fiberValidator.ctypes.data)


    def setPercentage(self, newPercentage):
        self.percentage = newPercentage


    def setStates(self, newStates):
        self.bundleStates = np.array(newStates, dtype=np.int32)



import numpy as np
from ...FiberVis_core import ROIsSegmentationQueryOctree, ROIsSegmentationPopulateAndDefragmentPool


class OctreePointBased:
    OT_NODE_CAPACITY = 256

    def __init__(self, points, fiberSizes):
        print('Building Octree...')
        self.vertex = points
        self.treeCount = 0
        self.root = 0
        self.pool = None
        self.vertexIdxPool = None

        x = self.vertex[0::3]
        y = self.vertex[1::3]
        z = self.vertex[2::3]

        xmin, xmax = x.min(), x.max()
        ymin, ymax = y.min(), y.max()
        zmin, zmax = z.min(), z.max()

        radius = np.array([(xmax-xmin)/2, (ymax-ymin)/2,
                          (zmax-zmin)/2], dtype=np.float32) * 1.005
        center = radius + np.array([xmin, ymin, zmin], dtype=np.float32)

        self.CPopulateAndDefragment(points.size, center, radius)

        self.createVertex2Fiber(fiberSizes)


    def CPopulateAndDefragment(self, vertexSize, center, radius):
        poolSize = int(vertexSize/self.OT_NODE_CAPACITY*4.4 + 1)

        dt = np.dtype([('children', np.uint32, (8,)),
                       ('numberOfElements', np.uint32, (1,)),
                       ('contentElements', np.uint32, (1,)),
                       ('radius', np.float32, (3,)),
                       ('center', np.float32, (3,))])

        self.pool = np.empty(poolSize, dtype=dt)
        self.vertexIdxPool = np.empty(self.vertex.size//3, dtype=np.uint32)

        n = ROIsSegmentationPopulateAndDefragmentPool(
            self.pool,
            poolSize,
            self.vertex,
            self.vertex.size,
            center,
            radius,
            self.vertexIdxPool)

        # n = cfuncs.ROIsSegmentationPopulateAndDefragmentPool(
        #       self.pool.ctypes.data,
        #       poolSize,
        #       self.vertex.ctypes.data,
        #       self.vertex.size,
        #       center.ctypes.data,
        #       radius.ctypes.data,
        #       self.vertexIdxPool.ctypes.data)

        self.pool = np.array(self.pool[:n], copy=True)

    def createVertex2Fiber(self, fiberSizes):
        self.vertexIdx2FiberIdx = np.empty(
            self.vertexIdxPool.size, dtype=np.int32)

        offset = 0
        for i in range(len(fiberSizes)):
            self.vertexIdx2FiberIdx[offset:offset+fiberSizes[i]] = i
            offset += fiberSizes[i]

    def queryCollision(self, queryPackage, result):
        ROIsSegmentationQueryOctree(len(queryPackage),
                                    queryPackage,
                                    self.pool,
                                    self.vertexIdxPool,
                                    self.vertexIdx2FiberIdx,
                                    self.vertex,
                                    result,
                                    result.shape[1])

        # cfuncs.ROIsSegmentationQueryOctree(   len(queryPackage),
        #                                                                       queryPackage.ctypes.data,
        #                                                                       self.pool.ctypes.data,
        #                                                                       self.vertexIdxPool.ctypes.data,
        #                                                                       self.vertexIdx2FiberIdx.ctypes.data,
        #                                                                       self.vertex.ctypes.data,
        #                                                                       result.ctypes.data,
        #                                                                       result.shape[1])



'''
Dictionary is keeping a reference to the segmentations objects... not deleting
'''

import PyQt5
from PyQt5 import QtGui, QtWidgets, uic, QtCore
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files
from .InPlaceSegmentationDialog import InPlaceSegmentationDialog
from .ROIsSegmentationDialog import ROIsSegmentationDialog
from .AtlasBasedParallelSegmentationDialog import AtlasBasedParallelSegmentationDialog
from .FFClustSegmentationDialog import FFClustSegmentationDialog
from ...Framework.Tools.visualizationEnums import SegmentationTypes, segmentations, mriVisualizations, VisualizationActions, VisualizationObject
from ...Framework.Segmentation.SegmentationHandler import SegmentationHandler
from ...Framework.Mesh import Mesh

_vot_ui = files('phybers.fibervis') / 'ui' / 'visualizationObjectsTool.ui'


def identifyNumberRecursively(item, number):
    root = item

    if item.parent() is not None:
        root = identifyNumberRecursively(item.parent(), number)
        number.append(item.parent().indexOfChild(item))
    else:
        return item
    return root


def retrieveItemIteratively(root, index):
    item = root

    for i in index:
        if hasattr(item, 'children'):
            item = item.children[i]
        else:
            item = item[i]

    return item


class ownQTreeWidgetItem(QtWidgets.QTreeWidgetItem):
    def setIdentifier(self, identification):
        self.identifier = identification


class VisDialog(QtWidgets.QDialog):
    ''' Window controller for the app. It has function for several action.
    '''

    closed = QtCore.pyqtSignal()
    changeObject = QtCore.pyqtSignal(dict)
    modifySegmentation = QtCore.pyqtSignal(dict)
    selectedObject = QtCore.pyqtSignal(dict)

    def __init__(self, parent):
        super().__init__(parent)
        self.ui = uic.loadUi(str(_vot_ui), self)

        self.ui.setWindowTitle('Visualization Tool')

        # signal when closing
        oldCloseEvent = self.ui.closeEvent

        def newCloseEvent(*args, **kwargs):
            self.closed.emit()
            return oldCloseEvent(*args, **kwargs)
        self.ui.closeEvent = newCloseEvent

        # this 4 variables have to be treated like const
        self.bundleReference = None
        self.meshReference = None
        self.mriReference = None
        self.roisReference = None

        # Context menu
        self.ui.treeWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.ui.treeWidget.customContextMenuRequested.connect(self.prepareMenu)

        # Setting widget
        # Signal for displaying settings
        self.ui.treeWidget.currentItemChanged.connect(self.activeItemChanged)

        self.connectAndConfigRotateGB()
        self.connectAndConfigTranslateGB()
        self.connectAndConfigScaleGB()
        self.ui.resetTransformsButton.clicked.connect(self.resetTransformsItemSelected)
        self.ui.loadMatrixFromFileButton.clicked.connect(self.loadMatrixFromFile)

        self.connectAndConfigBundleGB()
        self.connectAndConfigSliceGB()
        self.connectAndConfigVolumeGB()
        self.configMeshGB()

        # Hide rotate, translate and scale options, show when item is selected
        self.rotateGroupBox.hide()
        self.translateGroupBox.hide()
        self.scaleGroupBox.hide()

        self.bundleGroupBox.hide()
        self.sliceGroupBox.hide()
        self.meshGroupBox.hide()
        self.volumeGroupBox.hide()

        self.ui.resetTransformsButton.hide()

        # Setting windows
        self.windows = dict()

        # segmentation dialogs for identifier
        self.segmentationDialogs = {SegmentationTypes.InPlace : InPlaceSegmentationDialog,
                                                                SegmentationTypes.ROIs : ROIsSegmentationDialog,
                                                                SegmentationTypes.AtlasBased : AtlasBasedParallelSegmentationDialog,
                                                                SegmentationTypes.FFClust : FFClustSegmentationDialog}


    def connectAndConfigRotateGB(self):
        self.ui.rotateAngleLineEdit.returnPressed.connect(self.rotateItemSelected)
        self.ui.rotateXLineEdit.returnPressed.connect(self.rotateItemSelected)
        self.ui.rotateYLineEdit.returnPressed.connect(self.rotateItemSelected)
        self.ui.rotateZLineEdit.returnPressed.connect(self.rotateItemSelected)
        self.ui.rotateButton.clicked.connect(self.rotateItemSelected)

        validator = QtGui.QDoubleValidator(-1000, 1000, 2, self)
        self.ui.rotateAngleLineEdit.setValidator(validator)
        self.ui.rotateXLineEdit.setValidator(validator)
        self.ui.rotateYLineEdit.setValidator(validator)
        self.ui.rotateZLineEdit.setValidator(validator)


    def connectAndConfigTranslateGB(self):
        self.ui.translateXLineEdit.returnPressed.connect(self.translateItemSelected)
        self.ui.translateYLineEdit.returnPressed.connect(self.translateItemSelected)
        self.ui.translateZLineEdit.returnPressed.connect(self.translateItemSelected)
        self.ui.translateButton.clicked.connect(self.translateItemSelected)

        validator = QtGui.QDoubleValidator(-1000, 1000, 2, self)
        self.ui.translateXLineEdit.setValidator(validator)
        self.ui.translateYLineEdit.setValidator(validator)
        self.ui.translateZLineEdit.setValidator(validator)


    def connectAndConfigScaleGB(self):
        self.ui.scaleXLineEdit.returnPressed.connect(self.scaleItemSelected)
        self.ui.scaleYLineEdit.returnPressed.connect(self.scaleItemSelected)
        self.ui.scaleZLineEdit.returnPressed.connect(self.scaleItemSelected)
        self.ui.scaleButton.clicked.connect(self.scaleItemSelected)

        validator = QtGui.QDoubleValidator(-1000, 1000, 2, self)
        self.ui.scaleXLineEdit.setValidator(validator)
        self.ui.scaleYLineEdit.setValidator(validator)
        self.ui.scaleZLineEdit.setValidator(validator)


    def connectAndConfigBundleGB(self):
        self.bundleGBSignals()


    def connectAndConfigSliceGB(self):
        self.sliceGBSignals()

        validator = QtGui.QDoubleValidator(-1000, 1000, 2, self)
        self.ui.axisCXLE.setValidator(validator)
        self.ui.axisCYLE.setValidator(validator)
        self.ui.axisCZLE.setValidator(validator)


    def connectAndConfigVolumeGB(self):
        self.volumeGBSignals()


    def configMeshGB(self, action='connect'):
        getattr(self.ui.drawLinesCheckBox.clicked, action)(self.modifyMeshObject)
        getattr(self.ui.opacitySlider.valueChanged, action)(self.modifyMeshObject)
        getattr(self.ui.setColorButton.clicked, action)(self.modifyMeshObject)
        getattr(self.ui.front2backRadioButton.clicked, action)(self.modifyMeshObject)
        getattr(self.ui.back2frontRadioButton.clicked, action)(self.modifyMeshObject)


    def toggleVisWin(self, isChecked, pos, size):
        @QtCore.pyqtSlot()
        def toggle():

            if isChecked() == True:
                moveX = pos().x()-self.size().width()-5 # 5 pixels away from the window
                moveY = pos().y()-(self.size().height()-size().height())//2

                if moveX < 0:
                    moveX = 0
                if moveY < 0:
                    moveY = 0
                self.move(moveX,moveY)
                self.show()
            else:
                self.hide()
        return toggle


    def addRoot(self, name, ident):
        root = ownQTreeWidgetItem(self.ui.treeWidget)
        root.setText(0, name)
        self.ui.treeWidget.addTopLevelItem(root)
        root.setExpanded(True)
        root.setIdentifier(ident)
        return root


    def addChild(self, parent, name, ident):
        item = ownQTreeWidgetItem()
        item.setText(0, name)
        parent.addChild(item)
        item.setExpanded(True)
        item.setIdentifier(ident)
        return item


    def removeItem(self, item):
        if item.parent() == None:
            idx = self.ui.treeWidget.indexOfTopLevelItem(item)
            self.ui.treeWidget.takeTopLevelItem(idx)
        else:
            item.parent().removeChild(item)
        del item


    @QtCore.pyqtSlot()
    def updateTree(self):
        self.ui.treeWidget.clear()

        def recursivelyAddChildren(parentItem, parentModel):
            for child in (parentModel.children if hasattr(parentModel, 'children') else parentModel):
                childItem = self.addChild(parentItem, child.fileName, child.identifier)
                if child.identifier == VisualizationObject.Segmentation:
                    if not child in self.windows.keys():
                        try:
                            dialog = self.segmentationDialogs[child.segmentationIdentifier](child, self)
                            self.windows[child] = dialog
                            dialog.updateObject.connect(self.modifySegmentation)
                        except Exception as e:
                            print(e)
                            raise TypeError('Segmentation type dialog not implemented: ', child.segmentationIdentifier)

                    self.windows[child].update()

                recursivelyAddChildren(childItem, child)


        if len(self.bundleReference):
            bundleRoot = self.addRoot('Bundles', VisualizationObject.Bundle)

            recursivelyAddChildren(bundleRoot, self.bundleReference)

        if len(self.meshReference):
            meshRoot = self.addRoot('Meshes', VisualizationObject.Mesh)

            recursivelyAddChildren(meshRoot, self.meshReference)

        if len(self.mriReference):
            mriRoot = self.addRoot('MRIs', VisualizationObject.MRI)

            recursivelyAddChildren(mriRoot, self.mriReference)

        if len(self.roisReference):
            roisRoot = self.addRoot('ROIs', VisualizationObject.ROI)

            recursivelyAddChildren(roisRoot, self.roisReference)


    @QtCore.pyqtSlot(SegmentationHandler)
    def updateSettingWindow(self, segmentationRef):
        self.windows[segmentationRef].updateWindow()


    @QtCore.pyqtSlot(QtCore.QPoint)
    def prepareMenu(self, pos):
        item = self.ui.treeWidget.itemAt(pos)
        if item == None:
            return

        number = []
        root = identifyNumberRecursively(item,number)

        parent = item.parent()
        text = None
        if parent != None:
            text = parent.text(0)

        # Menu
        contextMenu = QtWidgets.QMenu(self.ui.treeWidget)

        # Actions
        deleteAction = contextMenu.addAction('Delete')
        addSegmentationAction = QtWidgets.QMenu('Add segmentation method')
        addMRIVisualizationAction = QtWidgets.QMenu('Add MRI visualization object')
        focusObjectAction = contextMenu.addAction('Focus Camera')
        visibleToggleAction = QtWidgets.QAction('Visible')
        visibleToggleAction.setCheckable(True)

        windowToggleAction = QtWidgets.QAction('Setting window')
        windowToggleAction.setCheckable(True)

        segmentationAction = dict()
        mriVisAction = dict()

        # Segmentation only on bundles
        if item.identifier == VisualizationObject.Bundle and parent != None:
            contextMenu.addMenu(addSegmentationAction)

            for key in segmentations.keys():
                segmentationAction[addSegmentationAction.addAction(key)] = segmentations[key]

        # MRI slicing and volume
        if item.identifier == VisualizationObject.MRI and parent != None:
            contextMenu.addMenu(addMRIVisualizationAction)

            for key in mriVisualizations.keys():
                mriVisAction[addMRIVisualizationAction.addAction(key)] = mriVisualizations[key]


        # Visible only on Objects
        if root.text(0) == 'Bundles':
            reference = self.bundleReference
        elif root.text(0) == 'Meshes':
            reference = self.meshReference
        elif root.text(0) == 'MRIs':
            reference = self.mriReference
        elif root.text(0) == 'ROIs':
            reference = self.roisReference
        else:
            reference = None

        if parent != None:
            visibleToggleAction.setChecked(retrieveItemIteratively(reference,number).drawable)
            contextMenu.addAction(visibleToggleAction)
        if item.identifier == VisualizationObject.Segmentation:
            dialog = self.windows[retrieveItemIteratively(reference,number)]
            windowToggleAction.setChecked(dialog.isVisible())
            contextMenu.addAction(windowToggleAction)

        action = contextMenu.exec_(self.ui.treeWidget.mapToGlobal(pos))

        modifyObject = None

        if action == deleteAction:
            if item.identifier == VisualizationObject.Segmentation:
                self.windows.pop(retrieveItemIteratively(reference,number), None)
            modifyObject = self.prepareDictionary(item, VisualizationActions.Delete)
        elif action == visibleToggleAction:
            modifyObject = self.prepareDictionary(item, VisualizationActions.ToggleDrawable)
        elif action == windowToggleAction:
            dialog = self.windows[retrieveItemIteratively(reference, number)]
            dialog.setVisible(not dialog.isVisible())
            return
        elif action in segmentationAction.keys():
            modifyObject = self.prepareDictionary(item, VisualizationActions.AddSegmentation, data=segmentationAction[action])
        elif action in mriVisAction.keys():
            modifyObject = self.prepareDictionary(item, VisualizationActions.AddSegmentation, data=mriVisAction[action])
        elif action == focusObjectAction:
            modifyObject = self.prepareDictionary(item, VisualizationActions.FocusObject)
        else:
            print('Action not atended: ', action)
            return

        self.changeObject.emit(modifyObject)


    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Delete:
            item = self.ui.treeWidget.currentItem()

            if item == None:
                return

            modifyObject = self.prepareDictionary(item, VisualizationActions.Delete)

            self.changeObject.emit(modifyObject)

        else:
            event.ignore()
            return

        event.accept()


    def identifyItem(self, item):
        if item.parent() == None:
            identifier = item.identifier
            number = -1
        else:
            number = []
            root = identifyNumberRecursively(item,number)
            identifier = root.identifier

        return identifier, number


    def prepareDictionary(self, item, action, data=None):
        opts = {}
        opts['rootType'], opts['number'] = self.identifyItem(item)
        opts['action'] = action

        if data != None:
            opts['data'] = data

        return opts


    @QtCore.pyqtSlot(QtWidgets.QTreeWidgetItem, QtWidgets.QTreeWidgetItem)
    def activeItemChanged(self, current, previous):
        if current == None:
            self.rotateGroupBox.hide()
            self.translateGroupBox.hide()
            self.scaleGroupBox.hide()
            self.ui.itemSelectedLineEdit.setText('None')

            self.bundleGroupBox.hide()
            self.sliceGroupBox.hide()
            self.meshGroupBox.hide()
            self.volumeGroupBox.hide()
            self.ui.resetTransformsButton.hide()

            activeObject = None

        else:
            self.rotateGroupBox.show()
            self.translateGroupBox.show()
            self.scaleGroupBox.show()
            self.ui.itemSelectedLineEdit.setText(current.text(0))
            self.ui.resetTransformsButton.show()

            number = []
            root = identifyNumberRecursively(current, number)

            # Visible only on Objects
            if root.text(0) == 'Bundles':
                reference = self.bundleReference
            elif root.text(0) == 'Meshes':
                reference = self.meshReference
            elif root.text(0) == 'MRIs':
                reference = self.mriReference
            elif root.text(0) == 'ROIs':
                reference = self.roisReference
            else:
                reference = None

            refObj = retrieveItemIteratively(reference, number)
            activeObject = refObj

            if current.identifier == VisualizationObject.Segmentation:
                print('SEGMENTATION')
            # --------------------BUNDLES--------------------
            if current.identifier == VisualizationObject.Bundle:
                self.bundleGroupBox.show()

                self.bundleGBSignals(action='disconnect')
                if isinstance(refObj, list):
                    self.ui.shaderRadioButton_0.setChecked(refObj[0].selectedShader == 0)
                    self.ui.shaderRadioButton_1.setChecked(refObj[0].selectedShader == 1)
                else:
                    self.ui.shaderRadioButton_0.setChecked(refObj.selectedShader == 0)
                    self.ui.shaderRadioButton_1.setChecked(refObj.selectedShader == 1)
                self.bundleGBSignals()

            else:
                self.bundleGroupBox.hide()

            # --------------------MRI SLICE--------------------
            if current.identifier == VisualizationObject.MRISlice:
                self.sliceGroupBox.show()

                self.sliceGBSignals(action='disconnect')
                self.ui.linearInterpCheckBox.setChecked(refObj.linearInterp)
                self.ui.discardPixelsCheckBox.setChecked(refObj.discardValues)
                self.ui.sliceThresholdSlider.setMaximum(int(refObj.max))
                self.ui.sliceThresholdSlider.setValue(int(refObj.threshold))
                self.ui.sliceCurrentThresholdLineEdit.setText(str(refObj.threshold))

                axis = refObj.axis == 1.0
                if axis.sum() == 1:
                    if axis[0] == 1.0:
                        self.ui.axisXRadioButton.setChecked(True)
                    elif axis[1] == 1.0:
                        self.ui.axisYRadioButton.setChecked(True)
                    elif axis[2] == 1.0:
                        self.ui.axisZRadioButton.setChecked(True)
                else:
                    self.ui.axisCustomRadioButton.setChecked(True)

                self.axisCXLE.setText(str(refObj.axis[0]))
                self.axisCYLE.setText(str(refObj.axis[1]))
                self.axisCZLE.setText(str(refObj.axis[2]))

                self.slicePos2AxisSlider.setValue(int(refObj.pos2Axis*(self.ui.slicePos2AxisSlider.maximum()+1)))

                self.sliceBrightSlider.setValue(int(refObj.bright*self.sliceBrightSlider.maximum()))
                self.sliceContrastSlider.setValue(int(refObj.contrast*10))
                self.sliceGBSignals()

            else:
                self.sliceGroupBox.hide()

            # --------------------MRI VOLUME--------------------
            if current.identifier == VisualizationObject.MRIVolume:
                self.volumeGBSignals(action='disconnect')
                self.volumeGroupBox.show()

                self.ui.volumeAlphaSlider.setValue(int(refObj.alpha*100))

                self.ui.volumeF2BCheckBox.setChecked(refObj.f2bDrawing)

                self.ui.volumeThresholdSlider.setMaximum(int(refObj.max)*100)
                self.ui.volumeThresholdSlider.setValue(int(refObj.threshold*100))
                self.ui.volumeCurrentThresholdLineEdit.setText(str(refObj.threshold))

                self.volumeGBSignals()

            else:
                self.volumeGroupBox.hide()


            # --------------------MESH--------------------
            if current.identifier == VisualizationObject.Mesh and len(number) != 0:
                assert isinstance(refObj, Mesh)
                self.meshGroupBox.show()

                self.configMeshGB(action='disconnect')
                self.ui.drawLinesCheckBox.setChecked(refObj.drawableLines)
                self.ui.opacitySlider.setValue(int(self.ui.opacitySlider.maximum()*refObj.alpha))

                self.ui.colorWidget.setAutoFillBackground(True)
                self.ui.colorWidget.setPalette(QtGui.QPalette(QtGui.QColor(*refObj.getRGBA256())))

                if refObj.back2frontSorting:
                    self.ui.back2frontRadioButton.setChecked(True)
                else:
                    self.ui.front2backRadioButton.setChecked(True)

                self.configMeshGB()

            else:
                self.meshGroupBox.hide()

        self.selectedObject.emit({'current' : activeObject})


    @QtCore.pyqtSlot()
    def rotateItemSelected(self):
        item = self.ui.treeWidget.currentItem()

        if item == None:
            return

        angle = self.ui.rotateAngleLineEdit.text()
        x = self.ui.rotateXLineEdit.text()
        y = self.ui.rotateYLineEdit.text()
        z = self.ui.rotateZLineEdit.text()

        angle = float(angle) if angle != '.' and len(angle) != 0 else 0.0
        x = float(x) if x != '.' and len(x) != 0 else 0.0
        y = float(y) if y != '.' and len(y) != 0 else 0.0
        z = float(z) if z != '.' and len(z) != 0 else 0.0

        axis = [x, y, z]

        if (x**2+y**2+z**2) < 0.0001:
            reply = QtWidgets.QMessageBox.question(self, "Invalid value", 'Axis for rotation was not given.', QtWidgets.QMessageBox.Ok)
            return

        modifyObject = self.prepareDictionary(item, VisualizationActions.Rotate, data=(angle, axis))

        self.changeObject.emit(modifyObject)


    @QtCore.pyqtSlot()
    def resetTransformsItemSelected(self):
        item = self.ui.treeWidget.currentItem()

        if item == None:
            return

        modifyObject = self.prepareDictionary(item, VisualizationActions.ResetTransforms)

        self.changeObject.emit(modifyObject)


    @QtCore.pyqtSlot()
    def loadMatrixFromFile(self):
        item = self.ui.treeWidget.currentItem()

        if item == None:
            return

        matrixFile, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select matrix", "", "matrix (*.npy *.trm *.txt)")

        if len(matrixFile) != 1:
            return

        modifyObject = self.prepareDictionary(item, VisualizationActions.LoadAndApplyMatrix, data=matrixFile[0])

        self.changeObject.emit(modifyObject)


    @QtCore.pyqtSlot()
    def translateItemSelected(self):
        item = self.ui.treeWidget.currentItem()

        if item == None:
            return

        x = self.ui.translateXLineEdit.text()
        y = self.ui.translateYLineEdit.text()
        z = self.ui.translateZLineEdit.text()

        x = float(x) if x != '.' and len(x) != 0 else 0.0
        y = float(y) if y != '.' and len(y) != 0 else 0.0
        z = float(z) if z != '.' and len(z) != 0 else 0.0

        v = [x, y, z]

        modifyObject = self.prepareDictionary(item, VisualizationActions.Translate, data=v)

        self.changeObject.emit(modifyObject)


    @QtCore.pyqtSlot()
    def scaleItemSelected(self):
        item = self.ui.treeWidget.currentItem()

        if item == None:
            print('SHOULDNT BE TRIGGERED')
            return

        x = self.ui.scaleXLineEdit.text()
        y = self.ui.scaleYLineEdit.text()
        z = self.ui.scaleZLineEdit.text()

        x = float(x) if x != '.' and len(x) != 0 else 1.0
        y = float(y) if y != '.' and len(y) != 0 else 1.0
        z = float(z) if z != '.' and len(z) != 0 else 1.0

        v = [x, y, z]

        modifyObject = self.prepareDictionary(item, VisualizationActions.Scale, data=v)

        self.changeObject.emit(modifyObject)


    @QtCore.pyqtSlot()
    def modifyBundleObject(self):
        item = self.ui.treeWidget.currentItem()
        if self.sender() in {self.ui.shaderRadioButton_0, self.ui.shaderRadioButton_1}:
            shaderSelected = 0 if self.ui.shaderRadioButton_0.isChecked() else 1
            action = VisualizationActions.ShaderSelection
            data = [shaderSelected]
        else:
            color = QtWidgets.QColorDialog.getColor()
            if not color.isValid():
                return
            action = VisualizationActions.ColorSelection
            data = [(color.red()/255, color.green()/255, color.blue()/255)]

        modifyObject = self.prepareDictionary(item, action, data=data)
        self.changeObject.emit(modifyObject)


    @QtCore.pyqtSlot()
    def modifySliceObject(self):
        item = self.ui.treeWidget.currentItem()

        # texture linear interpolation checkbox
        linearInterp = self.ui.linearInterpCheckBox.isChecked()

        # discard values under threshold
        discardValues = self.ui.discardPixelsCheckBox.isChecked()

        # threshold
        threshold = self.ui.sliceThresholdSlider.value()
        self.ui.sliceCurrentThresholdLineEdit.setText(str(threshold))

        # Axis
        if self.ui.axisCustomRadioButton.isChecked():
            x = self.ui.axisCXLE.text()
            y = self.ui.axisCYLE.text()
            z = self.ui.axisCZLE.text()

            x = float(x) if x != '.' and len(x) != 0 else 1.0
            y = float(y) if y != '.' and len(y) != 0 else 1.0
            z = float(z) if z != '.' and len(z) != 0 else 1.0

            if (x**2+y**2+z**2) < 0.0001:
                reply = QtWidgets.QMessageBox.question(self, "Invalid value", 'Axis for plane was not given.', QtWidgets.QMessageBox.Ok)
                return

            axis = [x, y, z]

        else:
            axis = [float(self.ui.axisXRadioButton.isChecked()),
                            float(self.ui.axisYRadioButton.isChecked()),
                            float(self.ui.axisZRadioButton.isChecked())]

        # pos2Axis slider
        pos2Axis = self.ui.slicePos2AxisSlider.value()/(self.ui.slicePos2AxisSlider.maximum()+1)

        # bright and contrast
        bright = self.ui.sliceBrightSlider.value()/self.ui.sliceBrightSlider.maximum()
        contrast = self.ui.sliceContrastSlider.value()/10

        modifyObject = self.prepareDictionary(item, VisualizationActions.SliceModification, data=(linearInterp, discardValues, threshold, axis, pos2Axis, bright, contrast))

        self.changeObject.emit(modifyObject)

    @QtCore.pyqtSlot()
    def modifyVolumeObject(self):
        item = self.ui.treeWidget.currentItem()

        if self.sender() == self.ui.volumeThresholdSlider or self.sender() == self.ui.volumeAlphaSlider:
            alpha = self.ui.volumeAlphaSlider.value()/100
            threshold = self.ui.volumeThresholdSlider.value()/100

            self.ui.volumeCurrentThresholdLineEdit.setText(str(threshold))

            data = [threshold, alpha]
        else:
            data = self.ui.volumeF2BCheckBox.isChecked()

        modifyObject = self.prepareDictionary(item, VisualizationActions.VolumeModification, data)

        self.changeObject.emit(modifyObject)



    @QtCore.pyqtSlot()
    def modifyMeshObject(self):
        item = self.ui.treeWidget.currentItem()

        sender = self.sender()

        if sender == self.ui.setColorButton:
            color = QtWidgets.QColorDialog.getColor()
            if color.isValid():
                self.ui.colorWidget.setPalette(QtGui.QPalette(color))
                color = (color.red()/255, color.green()/255, color.blue()/255)
            else:
                return

            data = [color]

        elif sender == self.ui.drawLinesCheckBox or sender == self.ui.opacitySlider:

            drawLines = self.ui.drawLinesCheckBox.isChecked()
            opacity = self.ui.opacitySlider.value()/self.ui.opacitySlider.maximum()

            data = (drawLines, opacity)

        elif sender == self.ui.front2backRadioButton:
            data = [False]

        elif sender == self.ui.back2frontRadioButton:
            data = [True]

        else:
            print('Unknown sender for modifyMeshObject in VisDialog:', self.sender())
            return

        modifyObject = self.prepareDictionary(item, VisualizationActions.MeshModification, data=data)

        self.changeObject.emit(modifyObject)


    @QtCore.pyqtSlot(list)
    def objectsChanged(self, objects):
        for dialogs in self.windows.keys():
            try:
                self.windows[dialogs].objectsChanged(objects)
            except:
                pass


    def bundleGBSignals(self, action='connect'):
        getattr(self.ui.shaderRadioButton_0.clicked, action) (self.modifyBundleObject)
        getattr(self.ui.shaderRadioButton_1.clicked, action) (self.modifyBundleObject)
        getattr(self.ui.changeBundleColorButton.clicked, action) (self.modifyBundleObject)

    def sliceGBSignals(self, action='connect'):
        getattr(self.ui.linearInterpCheckBox.clicked, action) (self.modifySliceObject)
        getattr(self.ui.discardPixelsCheckBox.clicked, action) (self.modifySliceObject)
        getattr(self.ui.sliceThresholdSlider.valueChanged, action) (self.modifySliceObject)
        getattr(self.ui.axisXRadioButton.clicked, action) (self.modifySliceObject)
        getattr(self.ui.axisYRadioButton.clicked, action) (self.modifySliceObject)
        getattr(self.ui.axisZRadioButton.clicked, action) (self.modifySliceObject)
        getattr(self.ui.axisCustomRadioButton.clicked, action) (self.modifySliceObject)
        getattr(self.ui.axisCXLE.returnPressed, action) (self.modifySliceObject)
        getattr(self.ui.axisCYLE.returnPressed, action) (self.modifySliceObject)
        getattr(self.ui.axisCZLE.returnPressed, action) (self.modifySliceObject)
        getattr(self.ui.slicePos2AxisSlider.valueChanged, action) (self.modifySliceObject)
        getattr(self.ui.sliceBrightSlider.valueChanged, action) (self.modifySliceObject)
        getattr(self.ui.sliceContrastSlider.valueChanged, action) (self.modifySliceObject)

    def volumeGBSignals(self, action='connect'):
        getattr(self.ui.volumeThresholdSlider.valueChanged, action) (self.modifyVolumeObject)
        getattr(self.ui.volumeAlphaSlider.valueChanged, action) (self.modifyVolumeObject)
        getattr(self.ui.volumeF2BCheckBox.clicked, action) (self.modifyVolumeObject)
        getattr(self.ui.volumeCurrentThresholdLineEdit.returnPressed, action) (self.modifyVolumeObject)



"""Module for finding and checking atlases."""
from pathlib import Path
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

_atlas_folder: Path = files('phybers.segment.fiberseg').joinpath('data')  # type: ignore
atlases = {k.stem: str(k) for k in _atlas_folder.glob('*')}


def check_atlas(filepath: str):
    """Check correct structure of atlas."""
    atlas_p = Path(filepath)
    info_p = atlas_p / "atlas_info_p.txt"
    bundles_p = atlas_p / "bundles"
    if info_p.is_file():
        names = set()
        with info_p.open() as f:
            names.add(f.readline().split()[0])
    else:
        return False
    if bundles_p.is_dir():
        found_names = set()
        for bundle in bundles_p.glob('*.bundles'):
            found_names.add(bundle.stem)
    else:
        return False
    if names != found_names:
        return False
    return True



from PyQt5 import QtGui, QtWidgets, uic, QtCore
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

_ffcsd_ui = files('phybers.fibervis.ui') / 'Segmentations' / 'FFClustSegmentationDialog.ui'

class FFClustSegmentationDialog(QtWidgets.QDialog):
    updateObject = QtCore.pyqtSignal(dict)


    def __init__(self, segmentation, parent):
        super().__init__(parent)
        self.ui = uic.loadUi(str(_ffcsd_ui), self)

        self.segmentationRef = segmentation


        self.ui.alphaSlider.valueChanged.connect(self.alphaChanged)

        self.ui.segmentButton.clicked.connect(self.segmentSubject)
        self.ui.exportBundleButton.clicked.connect(self.exportBundle)
        self.ui.exportCentroidButton.clicked.connect(self.exportCentroid)

        self.show()


    @QtCore.pyqtSlot()
    def segmentSubject(self):
        options = {     'reference' : self.segmentationRef,
                                'segmentSubject' : True}

        self.updateObject.emit(options)


    @QtCore.pyqtSlot()
    def alphaChanged(self):
        alpha = self.alphaSlider.value()/(self.alphaSlider.maximum()+1)

        options = {     'reference' : self.segmentationRef,
                                'attributes': {'setAlpha' : [alpha]}}

        self.updateObject.emit(options)


    @QtCore.pyqtSlot()
    def exportBundle(self):

        # Gets direction of bundle files
        outFile, fileType = QtWidgets.QFileDialog.getSaveFileName(self, "Save Bundle file", "", "Bundle (*.bundles)")#;;Trk file (*.trk)")

        options = {     'reference' : self.segmentationRef,
                                'attributes' : {'exportFile' : [outFile]}}

        self.updateObject.emit(options)


    @QtCore.pyqtSlot()
    def exportCentroid(self):
        # Gets direction of bundle files
        outFile, fileType = QtWidgets.QFileDialog.getSaveFileName(self, "Save Bundle file", "", "Bundle (*.bundles)")#;;Trk file (*.trk)")

        options = {     'reference' : self.segmentationRef,
                                'attributes' : {'exportCentroidFile' : [outFile]}}

        self.updateObject.emit(options)





    def __del__(self):
        print('Must clean befor leaving... INPLACESEGMENTATIONDIALOG', type(self))
        segmentationRef = None
        # super.__del__()


    def update(self):
        pass



from OpenGL import GL

class Shader:
    """ Shader class for GLSL programs
    """
    def __init__(self, vShader,  fShader, gShader=None):
        """
        Parameters
        ----------
        vShader : str or tuple
                String containing shader source code path (or tuple with paths)
                for the vertex shader
        fShader : str or tuple
                String containing shader source code path (or tuple with paths)
                for the fragment shader
        gShader : str or tuple
                String containing shader source code path (or tuple with paths)
                for the geometry shader

        """
        self.program = GL.glCreateProgram()
        shaders = []

        # vertex shaders
        if type(vShader) is list:
            shaders += [self._addShader(vertex, GL.GL_VERTEX_SHADER)
        for vertex in vShader]
        else:
            shaders.append(self._addShader(vShader, GL.GL_VERTEX_SHADER))

        # fragment shaders
        if type(fShader) is list:
            shaders += [self._addShader(fragment, GL.GL_FRAGMENT_SHADER)
        for fragment in fShader]
        else:
            shaders.append(self._addShader(fShader, GL.GL_FRAGMENT_SHADER))

        # geometry shaders
        if type(gShader) is list:
            shaders += [self._addShader(geometry, GL.GL_GEOMETRY_SHADER) for geometry in gShader]
        elif gShader is not None:
            shaders.append(self._addShader(gShader, GL.GL_GEOMETRY_SHADER))

        GL.glLinkProgram(self.program)

        # check if got linked
        if GL.glGetProgramiv(self.program, GL.GL_LINK_STATUS) != GL.GL_TRUE:
            info = GL.glGetProgramInfoLog(self.program)
            GL.glDeleteProgram(self.program)
            for shaders_id in shaders:
                GL.glDeleteShader(shaders_id)
            raise RuntimeError('Error linking program: %s' % (info))

        for shaders_id in shaders:
            GL.glDeleteShader(shaders_id)

    def _addShader(self, source_path, shader_type):
        """ Helper function for compiling a GLSL shader

        Parameters
        ----------
        source_path : str
                String containing shader source path

        shader_type : valid OpenGL shader type
                Type of shader to compile

        Returns
        -------
        value : int
                Identifier for shader if compilation is successful

        """
        shader_id = None
        try:
            shader_id = GL.glCreateShader(shader_type)
            with open(source_path, "r") as file:
                source = file.read()
                GL.glShaderSource(shader_id, source)
            GL.glCompileShader(shader_id)

            if GL.glGetShaderiv(shader_id, GL.GL_COMPILE_STATUS) != GL.GL_TRUE:
                info = GL.glGetShaderInfoLog(shader_id)
                raise RuntimeError('\n\tFile {1}:\n\tShader compilation failed:\n\t\t{0}'.format(info, source_path))

            GL.glAttachShader(self.program, shader_id)
            return shader_id
        except:
            if shader_id is not None:
                GL.glDeleteShader(shader_id)
            raise

    def glGetUniformLocation(self, name):
        """ Helper function to get location of an
        OpenGL uniform variable

        Parameters
        ----------
        name : str
                Name of the variable for which location
                is to be returned

        Returns
        -------
        value : int
                Integer describing location

        """
        uniform = GL.glGetUniformLocation(self.program, name)

        if uniform == -1:
            raise RuntimeError('Uniform not found: {0}'.format(name))
        else:
            return uniform

    def attributeLocation(self, name):
        """ Helper function to get location of an
        OpenGL attribute variable

        Parameters
        ----------
        name : str
                Name of the variable for which location is
                to be returned

        Returns
        -------
        value : int
                Integer describing location

        """
        attribute = GL.glGetAttribLocation(self.program, name)

        if attribute == -1:
            raise RuntimeError('Attribute not found: {0}'.format(name))
        else:
            return attribute

    def glUseProgram(self):
        """ Bipass for the function glUseProgram
        """

        GL.glUseProgram(self.program)



##
# @mainpage Hierarchical Clustering module
# @brief Brief HClust description.
# Descripción de HClust.
# \tableofcontents
# @subsection description_main How to run
# The hclust module is executed by running the following on console:
#
# > ```python3 -m hclust <dir_raw_tractography> <MaxDistance_Threshold> <maxdist> <var> <result_path>```
# Donde la definición de cada argumento se puede encontrar en el siguiente enlace: \ref hclust.main
# Where each argument's definition can be found in the following link: \ref hclust.main
#
#

import sys
from .mainHClust import hclust

if __name__ == "__main__":
    ## Path to the .bundles file that corresponds to the raw tractography.
    ## Ruta al archivo .bundles que corresponde a la tractografía.
    dir_raw_tractography = sys.argv[1]  # input format: ".bundles"

    ## Output directory path.
    ## Ruta del directorio de salida.
    result_path = sys.argv[2]

    #MaxDistance_Threshold="40" # variable threshold
    ## Maximum threshold of euclidian distance between pairs of fibers. 40 is recommended.
    ## Máximo umbral de distancia euclidiana entre pares de fibras. Se recomienda 40.
    MaxDistance_Threshold = int(sys.argv[3])  # variable threshold

    ## Maximum threshold of affinity, 30 is recommended.
    ## Máximo umbral de afinidad, se recomienda 30.
    maxdist = int(sys.argv[4])  # define usuario, se recomienda 30?

    ## Value related with the scale of similarity used. 3600 is recommended.
    ## Valor relacionado con la escala de similitud utilizada. Se recomienda 3600.
    var = int(sys.argv[5])  # define usuario, pero se recomienda usar 3600.


    hclust(dir_raw_tractography, result_path, MaxDistance_Threshold, maxdist, var)



# -*- coding: utf-8 -*-
"""Utils Module
"""
import os
import re
from shutil import rmtree
from typing import Tuple
from pathlib import Path
import numpy as np
import pandas as pd
from .FibersTools import getBundleSize, fiber_lens, Stadistics_txt_toxlsx
from .deform import deform as _deform
from .sampling import slice_fibers as _sampling
from .intersection import intersection as _intersection
from .postprocessing import postprocessing as _postprocessing


def check_input_path(in_file: str, ext="", dir=False):
    p = Path(in_file)
    if ext and ('.' not in p.name or p.name.rsplit('.', maxsplit=1)[1] != ext):
        raise ValueError(f"File {p.name} must have extension {ext}.")
    if dir and not p.is_dir():
        raise ValueError(f"Path {p} is not a directory.")
    if not p.exists():
        if dir:
            raise FileNotFoundError(f"Directory {p} not found.")
        else:
            raise FileNotFoundError(f"File {p} not found.")

def check_output_path(out_file: str, ext="", dir=False, overwrite=True):
    p = Path(out_file)
    if ext and ('.' not in p.name or p.name.rsplit('.', maxsplit=1)[1] != ext):
        raise ValueError(f"File {p.name} must have extension {ext}.")
    if dir:
        if p.exists():
            if not p.is_dir():
                raise ValueError(f"Path exist {p} but is not a directory.")
        else:
            os.makedirs(str(p))
    else:
        if p.exists() and not overwrite:
            raise ValueError(f"File {p} already exist and will not be deleted.")
        if p.is_dir():
            rmtree(str(p))
        if not p.parent.exists():
            os.makedirs(str(p.parent))



def deform(deform_file: str, file_in: str, file_out: str) -> None:
    """
    Transforms a tractography file to another space using a non-linear deformation image file.

    Parameters
    ----------
    deform_file : str
        Deformation image (image in NIfTI format containing the deformations).
    file_in : str
        Input tractography dataset in bundles format.
    file_out : str
        Path to the transformed tractography dataset.

    Returns
    -------
    None

    Notes
    -----
    This function generates the following files in the specified directory:

    Tractography : bundles file
        Tractography dataset that has been transformed into the MNI space.

    Examples
    --------
    To test `deform()`,  download the data from the `link deform <https://www.dropbox.com/sh/ncu8sf1ifwz4wpv/AACDzOXEdSrf8kBaWrbjfEPla?dl=1>`_.
    Then, open a Python terminal and run the following commands.

    >>> from phybers.utils import deform
    >>> deform( deform_file = 'defnolineal.nii', file_in = 'fibers.bundles', file_out = 'fibers_MNI.bundles' )

    Note: Make sure to replace the file paths with the actual paths to your data and directories.
    """
    check_input_path(deform_file, ext='nii')
    check_input_path(file_in)
    check_output_path(file_out, ext='bundles')
    _deform(deform_file, file_in, file_out)


def sampling(file_in: str, file_out: str, npoints: int = 21) -> None:
    """
    Performs a fiber sampling by recalculating their points using a specified number of equidistant points.

    Parameters
    ----------
    file_in : str
        Tractography dataset file in *'.bundles'* format.
    file_out : str
        Path to save the sub-sampled fibers.
    atlas_dir : str
        Bundle atlas, with bundles in separated files, sampled at 21 equidistant points. The bundle atlases provided are in same folders.
    npoints : str
        number of sampling points (n). Default: *21*.

    Returns
    -------
    None

    Notes
    -----
    This function generates the following files in the specified directory:

    Tractography : bundles file
        Tractography dataset sampled at n equidistant points.

    Examples
    --------
    To test `sampling()`,  download the data from the `link sampling <https://www.dropbox.com/sh/9tfxseo8uh68b32/AAAn56Xgiw7KhL2ILmkN6A23a?dl=1>`_.
    Then, open a Python terminal and run the following commands.

    >>> from phybers.utils import sampling
    >>> sampling ( file_in = 'sub_01.bundles', file_out = 'sub_01_21points.bundles', npoints = 21 )

    Note: Make sure to replace the file paths with the actual paths to your data and directories.
    """
    check_input_path(file_in, ext="bundles")
    check_output_path(file_out, ext="bundles")
    _sampling(file_in, file_out, npoints)


def inter2bundles(file1_in: str, file2_in: str, dir_out: str,
                  distance_thr: float = 10.0) -> Tuple[float, float]:
    """
    Calculates a similarity measure between two sets of brain fibers (fiber clusters or segmented bundles).
    The similarity measure yields a value between *0* and *100%*.

    Parameters
    ----------
    file1_in : str
        Path of the first fiber bundle.
    file2_in : str
        Path of the second fiber bundle.
    dir_out : str
        Path to save the distance matrix
    distance_thr : float
        Distance threshold in millimeters used to consider similar two fibers, default: *10*.

    Returns
    -------
    Tuple : [float, float]
        `[float, float]`, The first value indicates the percentage of intersection of the first set of fibers compared to the second set of fibers,
        and the second value indicates the reverse scenario, intersection of the second set of fibers compared to the first set of fibers.

    Examples
    --------
    To test `intersection()`, download the data from the `link intersection <https://www.dropbox.com/sh/dt196k65v03eh9m/AABKRW7ad2ssB0N_dpjqK4Dha?dl=1>`_.
    The provided data is identical, therefore a *100%* intersection is expected.
    Then, open a Python terminal and run the following commands.

    >>> from phybers.utils import intersection
    >>> result_inter=intersection ( file1_in = 'fib1.bundles', file2_in = 'fib2.bundles', dir_out = 'inter_result', distance_thr = 10.0 )
    >>> print(' intersection fibers1 with fibers2 ', result_inter [0] )
    >>> print(' intersection fibers2 with fibers1 ', result_inter [1] )

    Note: Make sure to replace the file paths with the actual paths to your data and directories.
    """


    labelsb1 = os.path.split(file1_in)[1]  # m - rows
    labelsb2 = os.path.split(file2_in)[1]  # num - columns


    p1 = 0
    p2 = 0

    str_out = labelsb1.split('.')[0]+'-'+labelsb2.split('.')[0]

    check_output_path(dir_out, dir=True)
    outfile = os.path.join(dir_out,str_out+'.txt')
    check_output_path(outfile)

    # computes the distance matrix between two fiber bundles.
    _intersection(file1_in, file2_in, outfile)

    ar=open(outfile,'rt')
    t=ar.readlines();ar.close()
    m=len(t)
    n=len(t[0].split('\t')[:-1])

    M = np.zeros((m,n),dtype = 'float32')

    for i in range(m):
        fila=t[i][:-1].split('\t')[:-1]
        for j in range(n):
            M[i,j]=float(fila[j])

    F_under_th=np.where( M < distance_thr) # fibers with at least 1 other fiber has distance under the threshold
    xp1=list(set(F_under_th[0]))
    yp2=list(set(F_under_th[1]))

    p1=(len(xp1)/getBundleSize(file1_in))*100 #fiber bundle 1 intersection percentage
    p2=(len(yp2)/getBundleSize(file2_in))*100 #fiber bundle 2 intersection percentage

    return (p1,p2)


def postprocessing(dir_in: str) -> pd.DataFrame:
    """
    Sets of algorithms that can be applied on the results of clustering and segmentation algorithms.

    Parameters
    ----------
    dir_in : str
        Root directory where all segmentation or clustering algorithm results are stored.

    Returns
    -------
    pd.DataFrame
        `pd.DataFrame`, contains the following list of keys: 'id_bundle': bundle identifier, 'sizes': number of fibers in the bundle,
        'lens': centroid length per bundle, 'intra_min': manimum intra-bundle Euclidean distance and intra_mean': mean intra-bundle Euclidean distance.

    Examples
    --------
    To test postprocessing(), download the data from the provided link, or execute it on the results of a clustering or segmentation algorithm.
    Then, open a Python terminal and run the following commands.


    >>> from phybers.utils import postprocessing
    >>> df = postprocessing ( dir_in = str )

    Note: Make sure to replace the file paths with the actual paths to your data and directories.
    """
    in_centroides=os.path.join(dir_in,'FinalCentroids','centroids.bundles')
    in_clusters_directory=os.path.join(dir_in,'FinalBundles')

    stadisticsFolder =  os.path.join(dir_in, 'stadistics')

    if not os.path.exists(dir_in):
        os.makedirs(dir_in)

    if not os.path.exists(stadisticsFolder):
        os.makedirs(stadisticsFolder)

    # Fibers Lens and Fibers Size
    txt_lens = ''
    txt_sizes = ''

#    stadisticsFolder =  os.path.join(dir_in, "stadistics")
    print(stadisticsFolder)

    if not os.path.exists(dir_in):
        os.makedirs(dir_in)
    if not os.path.exists(dir_in):
        os.makedirs(dir_in)
    centroides=read_bundle(in_centroides)
    for i in range(len(centroides)):
        txt_lens += str(fiber_lens(centroides[i]))+"\n"
        txt_sizes += str(getBundleSize(in_clusters_directory+'/'+str(i)+".bundles"))+"\n"

    with open(os.path.join(stadisticsFolder, 'lens.txt'), 'w') as file:
        file.write(txt_lens)

    with open(os.path.join(stadisticsFolder, 'sizes.txt'), 'w') as file:
        file.write(txt_sizes)

    # - Distancia intra-clúster mínima, Distancia intra promedio, Rij y DBindex

    _postprocessing(in_centroides, in_clusters_directory,
                    os.path.join(stadisticsFolder, 'mensure'), 21)

    # Dataframe
    Stadistics_txt_toxlsx(stadisticsFolder)

    return pd.read_excel(stadisticsFolder+'/stadistics.xlsx')


_p1 = re.compile(r'^.*\.bundles$')


def read_bundle(fp: str, npoints: int = 0) -> None:
    """
    Read bundles file.

    Parameters
    ----------
    file_in : str
        Tractography dataset file in *'.bundles'* format.
    npoints : str
        Path to save the sub-sampled fibers.


    Returns
    -------
    None

    Notes
    -----
    This function generates the following files in the specified directory:

    Tractography : bundles
        Tractography dataset sampled at n equidistant points.

    Examples
    --------
    To test `read_bundle()`,  download the data from the `link provided <https://www.dropbox.com/sh/9tfxseo8uh68b32/AAAn56Xgiw7KhL2ILmkN6A23a?dl=1>`_.
    Then, open a Python terminal and run the following commands.

    >>> from phybers.utils import sampling
    >>> sampling ( file_in = 'sub_01.bundles', file_out = 'sub_01_21points.bundles', npoints = 21 )

    Note: Make sure to replace the file paths with the actual paths to your data and directories.
    """
    if _p1.match(fp):
        fp += 'data'
    with open(fp, 'rb') as f:
        data = np.fromfile(f, dtype=np.float32)
    if npoints > 0:
        return data.reshape(-1, npoints * 3 + 1)[:, 1:].reshape(-1, npoints, 3)
    else:
        start = 0
        fibers = []
        size = data.size
        u32_data = data.view(np.uint32)
        while start < size:
            f_size = u32_data[start] * 3
            start += 1
            end = start + f_size
            fibers.append(data[start:end].reshape(-1, 3))
            start = end
        return fibers


_MINF_TEMPLATE = """attributes = {{
    'binary' : 1,
    'bundles' : {},
    'byte_order' : 'DCBA',
    'curves_count' : {},
    'data_file_name' : '*.bundlesdata',
    'format' : 'bundles_1.0',
    'space_dimension' : 3
}}
"""


def write_bundle(file_out: str, points, buffer_size=1_048_576)  -> None:
    """
    Write bundles file.

    Parameters
    ----------
    file_out : str
        Tractography dataset file in *'.bundles'* format.
    points : list
        Tractography dataset sampled at n equidistant points.

    Returns
    -------
    None

    Notes
    -----
    This function generates the following files in the specified directory:

    Tractography : bundles
        Tractography dataset sampled at n equidistant points.

    Examples
    --------
    To test `write_bundle()`,  download the data from the `link provided <https://www.dropbox.com/sh/9tfxseo8uh68b32/AAAn56Xgiw7KhL2ILmkN6A23a?dl=1>`_.
    Then, open a Python terminal and run the following commands.

    >>> from phybers.utils import sampling
    >>> sampling(dir_in='sub_01.bundles', file_out='sub_01_21points.bundles', npoints=21)

    Note: Make sure to replace the file paths with the actual paths to your data and directories.
    """
    if not _p1.match(file_out):
        file_out += '.bundles'
    if isinstance(points, np.ndarray) and points.ndim == 3:
        nfibers, npoints, _ = points.shape
        data = np.empty((nfibers, npoints * 3 + 1), np.float32)
        data.view(np.uint32)[:, 0] = npoints
        data[:, 1:] = points.reshape(-1, npoints * 3)
        with open(file_out + 'data', 'wb') as f:
            data.tofile(f)
    else:
        nfibers = len(points)
        buffer = bytearray(buffer_size * 4)
        u32_buffer = memoryview(buffer).cast('I')
        f32_buffer = memoryview(buffer).cast('f')
        start = 0
        with open(file_out + 'data', 'wb') as f:
            for fiber in points:
                assert isinstance(fiber, np.ndarray)
                end = 1 + fiber.size + start
                if end > buffer_size:
                    f.write(buffer[:start * 4])
                    start = 0
                    end = 1 + fiber.size + start
                u32_buffer[start] = fiber.shape[0]
                f32_buffer[start + 1:end] = fiber.flat[:]  # type: ignore
                start = end
            f.write(buffer[:start * 4])
    with open(file_out, 'wt') as f:
        f.write(_MINF_TEMPLATE.format(['points', 0], nfibers))



import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from PyQt5 import QtWidgets

parser = ArgumentParser(add_help=False, exit_on_error=False)
parser.add_argument('miscellaneous_files', nargs='*', default=(),
                    help='List of files or directories to be loaded '
                         'if supported.')
parser.add_argument('--bundles', nargs='*', default=(),
                    help='List of bundles files or directories containing '
                         'bundles files to be loaded.')
parser.add_argument('--mri', nargs='*', default=(),
                    help='List of mri files or directories containing '
                         'mri files to be loaded.')
parser.add_argument('--mesh', nargs='*', default=(),
                    help='List of mesh files or directories containing '
                         'mesh files to be loaded.')
parser.add_argument('--help', action='store_true')


def process_args(known_args: Namespace) -> dict[str, list[str]]:
    processed_paths: dict[str, list[str]] = {'bundles': [],
                                             'mri': [],
                                             'mesh': []}
    path_types = ('bundles', 'mri', 'mesh')
    path_formats = {'bundles': ('.bundles', '.trk', '.tck'),
                    'mri': ('.gz', '.nii'),
                    'mesh': ('.mesh', '.gii')}
    for path_type in path_types:
        for path_str in getattr(known_args, path_type):
            path = Path(path_str)
            if path.is_dir():
                for path_format in path_formats[path_type]:
                    processed_paths[path_type].extend(
                        map(str, path.glob(f'*{path_format}')))
            elif path.is_file():
                processed_paths[path_type].append(str(path))
            else:
                raise FileNotFoundError(path)
        if known_args.miscellaneous_files:
            for path_str in known_args.miscellaneous_files.copy():
                path = Path(path_str)
                if path.is_dir():
                    for path_format in path_formats[path_type]:
                        processed_paths[path_type].extend(
                            map(str, path.glob(f'*{path_format}')))
                elif path.is_file() and path.suffix in path_formats[path_type]:
                    known_args.miscellaneous_files.remove(path_str)
                    processed_paths[path_type].append(str(path))
                else:
                    known_args.miscellaneous_files.remove(path_str)
                    print(f'Path {path} not found, skiping.')
    return processed_paths


def parse_args_and_run():
    namespace, args = parser.parse_known_args(sys.argv[1:])
    app = QtWidgets.QApplication(args)
    if namespace.help:
        message_box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Information,
                                            "Help", f"<pre>{parser.format_help()}</pre>")
        message_box.exec()
    else:
        from .WindowController import WindowController

        paths = process_args(namespace)
        fiber_app = WindowController()
        fiber_app.show()
        fiber_app.contextHandler.addBundleFile(paths['bundles'])
        fiber_app.contextHandler.addMRIFile(paths['mri'])
        fiber_app.contextHandler.addMeshFile(paths['mesh'])
        app.exec()



def main(bundles=(), mri=(), mesh=(), args: list[str] = []):
    
    """
    Initializes the graphical user interface (GUI).
    
    Parameters
    ----------
    None

    Examples
    --------
    To test `fibervis()`, download the data from the links provided above. Then, open a Python terminal and run the following commands:

    .. code:: python

       from phybers.fibervis import start_fibervis
       start_fibervis()
    
    `fibervis()` is installed as a program, allowing you to run it through the command line in Windows or Ubuntu. To execute it on both platforms, use the following command:

    .. code:: python

       fibervis

    For your convenience in using `fibervis()`, a video demonstrating all its features is accessible through the following link:
    :video_link: [Video Link][video].    
    """

    from .WindowController import WindowController

    app = QtWidgets.QApplication(args)
    fiber = WindowController()
    fiber.show()
    fiber.contextHandler.addBundleFile(bundles)
    fiber.contextHandler.addMRIFile(mri)
    fiber.contextHandler.addMeshFile(mesh)
    return app.exec_()



import time

def timeit(func):

    def timed(*args, **kw):
        ts = time.time()
        result = func(*args, **kw)
        te = time.time()

        print ('%r (%r, %r) %2.4f sec' % (func.__name__, args, kw, te-ts))
        return result

    return timed

def averageTimeit(func):
    n=50

    def timed(*args, **kw):
        time_average = 0
        for i in range(n):
            print('test n: ', i)
            ts = time.time()
            result = func(*args, **kw)
            te = time.time()

            time_average += te-ts

        time_average /= n


        print ('%r (%r, %r) %d %2.4f sec' % (func.__name__, args, kw, n, time_average))
        return result

    return timed



import numpy as np

def bytes2MB(bytes):
    return bytes/1024

def bytes2GB(bytes):
    return bytes/(1024*1024)

def findIntegersMultiplierFor(n, under=1):
    for i in range(under):
        tmpN = n+i
        m1 = 1
        m2 = n

        while m1<=m2:
            if tmpN%m1 == 0 and m2<=under:
                return (np.int32(m2), np.int32(m1))
            m1 += 1
            m2 = tmpN//m1

    return None



from pathlib import Path
from os.path import exists, getmtime
from setuptools.command.build_ext import new_compiler  # type: ignore
from setuptools import setup, Extension, find_packages
import numpy

long_description = (Path(__file__).parent / "README.md").read_text()


if new_compiler().compiler_type == "msvc":
    openmp_compile = ["/openmp"]
    openmp_link = []
    fast_math = ["/fp:fast"]
else:
    openmp_compile = ["-fopenmp"]
    openmp_link = ["-fopenmp"]
    fast_math = ["-ffast-math"]

try:
    from Cython.Compiler import Options
    from Cython.Build import cythonize
    Options.fast_fail = True
    cython_ok = True
except ImportError:
    cython_ok = False


def is_up_to_date(c_path, cy_path):
    return getmtime(c_path) >= getmtime(cy_path)


def check_src(c_path):
    base, _ = c_path.rsplit('.', maxsplit=1)
    cy_path = base + ".pyx"
    if exists(c_path):
        if exists(cy_path):
            if is_up_to_date(c_path, cy_path):
                pass
            elif cython_ok:
                cythonize(cy_path)
            else:
                pass # Warning(f"Cython source {cy_path} is more recente than {language} source, but Cython is not present to update it")
    elif cython_ok:
        if exists(cy_path):
            cythonize(cy_path)
        else:
            raise RuntimeError(f"Both source file {c_path} and cython source file {cy_path} are missing.")
    else:
        raise RuntimeError(f"Source file {c_path} missing and Cython is not present.")
    return c_path


ffclust_ext = Extension(
    "phybers.clustering.ffclust.c_wrappers",
    sources=[
        check_src("./src/phybers/clustering/ffclust/c_wrappers.c"),
        "./src/phybers/clustering/ffclust/segmentation.c",
        "./src/phybers/clustering/ffclust/BundleTools.c"
    ],
    extra_compile_args=openmp_compile + fast_math,
    extra_link_args=openmp_link
)

hclust_ext = Extension(
    "phybers.clustering.hclust.c_wrappers",
    sources=[
        check_src("./src/phybers/clustering/hclust/c_wrappers.cpp"),
        "./src/phybers/clustering/hclust/distances.cpp",
        "./src/phybers/clustering/hclust/BundleTools.cpp"
    ]
)

segmentation_ext = Extension(
    "phybers.segment.fiberseg.c_wrappers",
    sources=[
        check_src("./src/phybers/segment/fiberseg/c_wrappers.cpp"),
        "./src/phybers/segment/fiberseg/main.cpp",
        "./src/phybers/segment/fiberseg/BundleTools.cpp"
    ],
    extra_compile_args=openmp_compile + fast_math,
    extra_link_args=openmp_link
)

deformHCPtoMNI_ext = Extension(
    "phybers.utils.deform.c_wrappers",
    sources=[
        check_src("./src/phybers/utils/deform/c_wrappers.c"),
        "./src/phybers/utils/deform/main.c",
        "./src/phybers/utils/deform/BundleTools.c"
    ],
)

fiberDistanceMax2bun_ext = Extension(
    "phybers.utils.intersection.c_wrappers",
    sources=[
        check_src("./src/phybers/utils/intersection/c_wrappers.c"),
        "./src/phybers/utils/intersection/fiberDistanceMax2bun.c",
        "./src/phybers/utils/intersection/BundleTools.c"
    ],
)

sliceFibers_ext = Extension(
    "phybers.utils.sampling.c_wrappers",
    sources=[
        check_src("./src/phybers/utils/sampling/c_wrappers.c"),
        "./src/phybers/utils/sampling/sliceFibers.c",
        "./src/phybers/utils/sampling/BundleTools.c"
    ],
)

postprocessing_ext = Extension(
    "phybers.utils.postprocessing.c_wrappers",
    sources=[
        check_src("./src/phybers/utils/postprocessing/c_wrappers.cpp"),
        "./src/phybers/utils/postprocessing/dbindex.cpp",
        "./src/phybers/utils/postprocessing/BundleTools.cpp"
    ],
    extra_compile_args=openmp_compile + fast_math,
    extra_link_args=openmp_link
)

fibervis_ext = Extension("phybers.fibervis.FiberVis_core.FiberVis_core",
                         sources=["./src/phybers/fibervis/FiberVis_core/FiberVis_core.c",
                                  "./src/phybers/fibervis/FiberVis_core/bundleMethods.c",
                                  "./src/phybers/fibervis/FiberVis_core/PointOctree.c",
                                  "./src/phybers/fibervis/FiberVis_core/AtlasBasedParallelSegmentation.c",
                                  "./src/phybers/fibervis/FiberVis_core/miscellaneous.c"],
                         include_dirs=[numpy.get_include()],
                         extra_compile_args=openmp_compile,
                         extra_link_args=openmp_link
)

setup(
    name="phybers",
    version="1.1.2b0",
    description="Phybers: A Package for Brain Tractography Analysis.",
    url="https://github.com/phybers/phybers",
    author="L. Liset Gonzalez, Alejandro Cofre",
    author_email="phybers.dmris@gmail.com",
    license="GNU",
    install_requires=[
        "numpy",
        "dipy",
        "joblib",
        "matplotlib",
        "scikit-learn",
        "networkx",
        "pandas",
        "PyQt5",
        "PyOpenGL",
        "PyOpenGL_accelerate",
        "nibabel",
        "pydicom",
        "scikit-image",
        "scipy",
        "openpyxl",
        "regex",
        "importlib_resources"
    ],
    package_dir={'': "src"},
    packages=find_packages(where="./src",
                           exclude=['phybers.fibervis_backup',
                                    'phybers.phyexample']),
    ext_modules=[
        ffclust_ext,
        hclust_ext,
        segmentation_ext,
        deformHCPtoMNI_ext,
        fiberDistanceMax2bun_ext,
        sliceFibers_ext,
        postprocessing_ext,
        fibervis_ext
    ],
    include_package_data=True,
    package_data={
        "phybers.fibervis": ["shaders/*"],
        "phybers.fibervis.ui": ["Segmentations/*", "Settings/*", "*.ui"],
        "phybers.clustering.ffclust": ["*.jpg"],
        "phybers.segment.fiberseg": ["data/*", "*.bundles",
                                     "*.bundlesdata", "*.txt"],
        "phybers": ["phyexample/*"]
    },
    exclude_package_data={'phybers.clustering': ["*.c", "*.cpp",
                                                 "*.h", "*.hpp"]},
    include_dirs=[numpy.get_include()],
    zip_safe=False,
    entry_points={
        'gui_scripts': [
            'fibervis = phybers.fibervis.FiberVis:parse_args_and_run'
        ],
    },
    long_description=long_description,
    long_description_content_type='text/markdown'
)



''' _calculateNormals ponderarlo por areas.
_loadMeshAttributes sin implementar (leia archivos .mesh.minf)
'''

import nibabel as nib
import numpy as np
import ctypes as ct
from pathlib import Path
from OpenGL import GL
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files
from .VisualizationBaseObject import VisualizationBaseObject, fSize, propagateToChildren, drawable, config
from .Tools.visualizationEnums import VisualizationObject
from .Shaders import Shader
from .BoundingBox import BoundingBox

_m_vs = files('phybers.fibervis') / 'shaders' / 'mesh.vs'
_sfs_fs = files('phybers.fibervis') / 'shaders' / 'standardFragmentShader.fs'


class Mesh(VisualizationBaseObject):
    ''' Class for drawing mesh files
    '''

    def __init__(self, sPath, shaderDict, parent):
        ''' Initialize a Mesh object.
        It copies references to the path and the GL.gl program.
        Then it loads the information from the file (points, normals, faces).

        Sends the information to the GPU buffers.

        Initialize  all parameters.

        The drawable variable is set to True.

        Parameters
        ----------
        sPath : str
            String containing path for the mesh file (.mesh)
        shaderProgram : GL.GLint
            Reference to the GL.gl program containing all the shaders

        '''

        super().__init__(parent, shaderDict)
        self.identifier = VisualizationObject.Mesh

        self.path = sPath
        self.fileName = Path(sPath).name

        self.points = None
        self.normals = None
        self.faces = None

        # Read the vertex and connectivity for the mesh
        self._readMesh()

        # Creates VBOs & an EBO, then populates them with the point and normal data. The elements data goes into the EBO
        self._loadBuffers()

        self.triangleColor = np.array([0.7, 0.7, 0.7], dtype=np.float32)
        self.lineColor = np.array([0.0, 0.0, 0.0], dtype=np.float32)
        self.alpha = 0.5

        # Uniform location
        self.colorUniform = self.shader[self.selectedShader].glGetUniformLocation("meshColor")

        # Ready to draw
        self.drawable = True
        self.drawableLines = False

        # We set the boundingbox parameters
        x = self.points[:,0]
        y = self.points[:,1]
        z = self.points[:,2]

        xmin, xmax = x.min(),x.max()
        ymin, ymax = y.min(),y.max()
        zmin, zmax = z.min(),z.max()

        dims = np.array([xmax-xmin, ymax-ymin, zmax-zmin], dtype=np.float32)
        center = np.array([xmin, ymin, zmin], dtype=np.float32)

        self.boundingbox = BoundingBox(shaderDict, self, dims, center)

        self.back2frontSorting = True

        print("Loading ready:\n\t", self.faces.shape[0], " triangles.\n\t", self.points.shape[0], " vertex.")

        # Not effi
        self.calculateTriangleCentroid()


    def cleanOpenGL(self):
        print('Cleaning object: ', self)
        GL.glDeleteVertexArrays(1, [self.vao])

        GL.glDeleteBuffers(1, [self.vbo])
        GL.glDeleteBuffers(1, [self.ebo])

        self.clean = True

    def get_center(self):
        return self.boundingbox.get_center() + self.translateMat[:3, 3].flat

    def get_size(self):
        return self.boundingbox.get_size()

    def _readMesh(self):
        ''' It reads the information from self.path file.
        Also calculates the normals.
        Finally creates the ebo buffer.

        Allowed files: .mesh

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        extension = Path(self.path).suffix

        # Bundle file
        if extension == '.mesh':
            self._openMesh()

        elif extension == '.gii':
            self._openGIFTI()

        # Unsopported file
        else:
            raise TypeError('Unsupported file.')


    def _openMesh(self):
        ''' This function reads vertex, faces and (if in file) normals.
        It then creates the numpy array for points, normals and ebo (elements).

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        print('Opening mesh file {}'.format(self.path))

        # This was programed with meshes with triangles an not tetrahedron in mind
        with open(self.path, "rb") as file:
            # We read the endianess and prepare endianess variables
            endian = 'little' if file.read(9) == b'binarDCBA' else 'big'
            dtype = '<'  if endian == 'little' else '>'

            textureTypeU32 = int.from_bytes(file.read(4), byteorder=endian)
            textureType = file.read(4)

            polygonDimension = int.from_bytes(file.read(4), byteorder=endian)
            numberOfTimeSteps = int.from_bytes(file.read(4), byteorder=endian)

            print('endian: ', endian, '\ntextureTypeU32: ', textureTypeU32, '\ntextureType: ', textureType, '\npolygonDimension: ', polygonDimension, '\nnumberOfTimeSteps', numberOfTimeSteps)

            vertexVectorTotal = []
            normalVectorTotal = []
            polyVectorTotal = []

            vertexVector = []
            normalVector = []
            polyVector = []

            for timeStep in range(numberOfTimeSteps):
                # Read instant to be read (unsigned int 32 bits)
                actualTS = int.from_bytes(file.read(4), byteorder=endian)
                print('instant: ', actualTS)
                if actualTS != timeStep:
                    raise RuntimeError('Error with mesh file {0}. Current instant \'{1}\' does not match with instant readed: \'{2}\''.format(self.path, timeStep, actualTS))

                # Read number of vertices
                vertexVectorTotal.append(int.from_bytes(file.read(4), byteorder=endian))
                print('vertex_vector_total: ', vertexVectorTotal[-1])

                vertex = np.fromstring(file.read(4*vertexVectorTotal[-1]*polygonDimension), dtype=dtype+'f4')
                vertex = vertex.reshape((vertexVectorTotal[-1], polygonDimension))
                vertexVector.append(vertex)

                # Read number of normals
                normalVectorTotal.append(int.from_bytes(file.read(4), byteorder=endian))
                print('normal_vector_total: ', normalVectorTotal[-1])

                if normalVectorTotal[-1] == vertexVectorTotal[-1]:
                    normals = np.fromstring(file.read(4*normalVectorTotal[-1]*polygonDimension), dtype=dtype+'f4')
                    normals = normals.reshape(normalVectorTotal[-1], polygonDimension)
                    normalVector.append(normals)

                # Read the texture coordinates, if different from 0 raise an error
                textureVectorTotal = int.from_bytes(file.read(4), byteorder=endian)
                print('texture_vector_total: ', textureVectorTotal)

                if textureVectorTotal != 0:
                    raise RuntimeError('Mesh files with texture coordinates can not be read. Not implemented (./Framework/Mesh.py)')

                polyVectorTotal.append(int.from_bytes(file.read(4), byteorder=endian))
                print('poly_vector_total: ', polyVectorTotal[-1])

                poly = np.fromstring(file.read(4*polyVectorTotal[-1]*polygonDimension), dtype=dtype+'u4')
                poly = poly.reshape(polyVectorTotal[-1], polygonDimension)
                polyVector.append(poly)

                # If normals havent been given, we net to calculate them
                if normalVectorTotal[-1] != vertexVectorTotal[-1]:
                    normalVectorTotal[-1] = vertexVectorTotal[-1]
                    normals = Mesh._calculateNormals(vertexVector[-1], polyVector[-1])
                    normalVector.append(normals)

            # This will raise an error when a mesh with more than 1 time step is read
            self.points = np.array(vertexVector).reshape(vertexVector[0].shape)
            self.normals = np.array(normalVector).reshape(normalVector[0].shape)
            self.faces = np.array(polyVector).reshape(polyVector[0].shape)


    def _openGIFTI(self):
        ''' This function reads vertex, faces and (if in file) normals.
        It then creates the numpy array for points, normals and ebo (elements).

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''
        print('Opening gifti file {}'.format(self.path))

        giftiMesh = nib.load(self.path)

        if (len(giftiMesh.darrays) != 2):
            print('!!!!!!!! darrays = ', len(giftiMesh.darrays))

        self.points = giftiMesh.darrays[0].data
        self.faces = giftiMesh.darrays[1].data

        # print(type(self.points))
        # print(type(self.faces))

        # nib.freesurfer.io.write_geometry("test.gii", self.faces, self.points, create_stamp=None, volume_info=None)
        self.normals = Mesh._calculateNormals(self.points, self.faces)

    @staticmethod
    def _calculateNormals(vertex, triangles):
        ''' Anonymous function.
        It calculates the normal for each triangle and then calculates the average per vertex.
        '''

        normals = np.zeros(vertex.shape, dtype=np.float32)

        for triangle in triangles:
            # n = np.cross(vertex[triangle[1]]-vertex[triangle[0]],     vertex[triangle[2]]-vertex[triangle[0]])
            # n = n/np.linalg.norm(n)

            # normals[triangle] += n
            normals[triangle] += np.cross(vertex[triangle[1]]-vertex[triangle[0]],      vertex[triangle[2]]-vertex[triangle[0]])

        normals /= np.linalg.norm(normals, axis=1).reshape(normals.shape[0],1)

        return normals


    def _loadBuffers(self):
        ''' It generates a VAO, VBO and EBO.
        It populates them with the vertex and normal (the VBO), also the faces (the EBO).

        It finishes when the buffers are linked with attributes on the vertex shader.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''

        # Reference for vbos, ebos and attribute ptrs
        self.vao = GL.glGenVertexArrays(1)
        GL.glBindVertexArray(self.vao)

        self.vbo = GL.glGenBuffers(1)
        self.ebo = GL.glGenBuffers(1)

        # VBO
        bufferSize = self.points.nbytes + self.normals.nbytes
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.vbo)
        GL.glBufferData(GL.GL_ARRAY_BUFFER, bufferSize, None, GL.GL_STATIC_DRAW)        # Create empty buffer

        # Populate buffer with points
        GL.glBufferSubData(GL.GL_ARRAY_BUFFER,
            0,
            self.points.nbytes,
            self.points.flatten())

        # Normals
        GL.glBufferSubData(GL.GL_ARRAY_BUFFER,
            self.points.nbytes,
            self.normals.nbytes,
            self.normals.flatten())

        # EBO
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.ebo)
        GL.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, self.faces.nbytes, None, GL.GL_STATIC_DRAW)

        # Enable attributes
        positionAttribute =     self.shader[0].attributeLocation('vertexPos')
        normalAttribute =       self.shader[0].attributeLocation('vertexNor')

        # # Connect attributes
        GL.glEnableVertexAttribArray(positionAttribute)
        GL.glVertexAttribPointer(positionAttribute,3, GL.GL_FLOAT, GL.GL_FALSE, 3*fSize, None)

        GL.glEnableVertexAttribArray(normalAttribute)
        GL.glVertexAttribPointer(normalAttribute,       3, GL.GL_FLOAT, GL.GL_FALSE, 3*fSize, ct.c_void_p(self.points.nbytes))


    def _loadMeshAttributes(self):
        ''' Not implemented
        '''

        ns = dict()
        exec(open(self.path + '.minf').read(), ns)
        self.attributes = ns['attributes']

        item = 0
        for i in self.attributes['transformations']:
            print(item)
            print(i)
            item += 1


    def setDrawLines(self, drawLines):
        self.drawableLines = drawLines


    def setAlpha(self, newAlpha):
        self.alpha = newAlpha


    def setColor(self, color):
        self.triangleColor[0] = color[0]
        self.triangleColor[1] = color[1]
        self.triangleColor[2] = color[2]


    def getRGB256(self):
        return self.triangleColor*255

    def getRGBA256(self):
        r = int(self.triangleColor[0])
        g = int(self.triangleColor[1])
        b = int(self.triangleColor[2])
        a = int(self.alpha * 255)
        return  r, g, b, a


    @propagateToChildren
    @drawable
    @config
    def draw(self, eye):
        '''  It has 2 decorators configuring the vao and seeing if the visualize flag is up.

        It calls the draw element.
        '''
        self.sortTriangles(eye, self.back2frontSorting)
        self.loadSortedEBO()
        opacityUniform = self.shader[self.selectedShader].glGetUniformLocation('opacity')

        # Draw lines
        if self.drawableLines:

            GL.glUniform1f(opacityUniform, 1.0)
            GL.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE)
            GL.glUniform3fv(self.colorUniform, 1, self.lineColor)
            GL.glLineWidth(0.01)
            GL.glDrawElements(GL.GL_TRIANGLES, self.faces.size, GL.GL_UNSIGNED_INT, None)
            GL.glLineWidth(1.0)
            GL.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_FILL)

        # Draw triangles
        # GL.glDepthMask(GL.GL_FALSE)
        GL.glEnable(GL.GL_POLYGON_OFFSET_FILL)
        GL.glEnable(GL.GL_BLEND)
        GL.glEnable(GL.GL_CULL_FACE)

        GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
        GL.glPolygonOffset(1.0, 1.0)
        GL.glUniform1f(opacityUniform, self.alpha)
        GL.glUniform3fv(self.colorUniform, 1, self.triangleColor)

        GL.glDrawElements(GL.GL_TRIANGLES, self.faces.size, GL.GL_UNSIGNED_INT, None)

        GL.glDisable(GL.GL_POLYGON_OFFSET_FILL)
        GL.glDisable(GL.GL_BLEND)
        GL.glDisable(GL.GL_CULL_FACE)
        # GL.glDepthMask(GL.GL_TRUE)

        self.boundingbox.draw()

    @staticmethod
    def createProgram():
        ''' Anonymous function.
        It creates the shader programs for this specific class and returns the handler.
        '''

        vertexGLSL = [str(_m_vs)]
        fragmentGLSL = [str(_sfs_fs)]
        return [Shader(vertexGLSL, fragmentGLSL)]




    ## Testing concept
    def calculateTriangleCentroid(self):
        self.triangleCentroid = np.zeros(self.faces.shape, dtype=np.float32)

        for index in range(self.faces.shape[0]):
            self.triangleCentroid[index] = self.points[self.faces[index]].sum(axis=0)/3


    def sortTriangles(self, eye, back2front=True):
        relativeEye = (self.inverseModel * np.concatenate((eye, [1.0])).reshape((4,1)))[:3].ravel()

        tmp = np.copy(self.triangleCentroid)
        tmp -= relativeEye
        tmp = np.linalg.norm(tmp, axis=1)

        self.sorted = np.argsort(tmp)[::-1] if back2front == True else np.argsort(tmp)


    def loadSortedEBO(self):
        GL.glBindVertexArray(self.vao)

        GL.glBufferSubData(GL.GL_ELEMENT_ARRAY_BUFFER,
            0,
            self.faces.nbytes,
            self.faces[self.sorted].flatten())

    def setFront2BackSorting(self, flag):
        self.back2frontSorting = flag

    @staticmethod
    def validExtension():
        return ['mesh', 'gii']



from PyQt5 import QtGui, QtWidgets, uic, QtCore
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

_abpsd_ui = files('phybers.fibervis.ui') / 'Segmentations' / 'AtlasBasedParallelSegmentationDialog.ui'

class AtlasBasedParallelSegmentationDialog(QtWidgets.QDialog):
    updateObject = QtCore.pyqtSignal(dict)

    def __init__(self, segmentation, parent):
        super().__init__(parent)
        self.ui = uic.loadUi(str(_abpsd_ui), self)

        self.segmentationRef = segmentation

        self.bundleNames = None
        self.bundleSize = None
        self.states = []

        self.ui.tableWidget.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.ui.tableWidget.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.ui.tableWidget.setColumnWidth(0, 150)
        self.ui.tableWidget.setColumnWidth(2, 50)
        self.ui.tableWidget.horizontalHeader().setStretchLastSection(True)

        self.ui.tableWidget.cellClicked.connect(self.toggleCheck)

        self.ui.checkSegmentAll.clicked.connect(self.checkSegmentAllBundles)

        self.ui.alphaSlider.valueChanged.connect(self.alphaChanged)

        self.ui.addBundlesButton.clicked.connect(self.addBundles)
        self.ui.removeBundlesButton.clicked.connect(self.removeBundles)
        self.ui.applyMatrixButton.clicked.connect(self.applyMatrix)

        self.ui.addThresholdFileButton.clicked.connect(self.addThresholdFile)

        self.ui.filterLineEdit.textChanged.connect(self.filter)

        self.ui.segmentButton.clicked.connect(self.segmentSubject)
        self.ui.exportBundleButton.clicked.connect(self.exportBundle)

        self.show()


    @QtCore.pyqtSlot()
    def segmentSubject(self):
        options = {     'reference' : self.segmentationRef,
                                'segmentSubject' : True}

        self.updateObject.emit(options)


    def checkSegmentAllBundles(self):
        print('checkSegmentAll')


    def addThresholdFile(self):
        # Gets direction of bundle files
        thresholdFile, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select threshold file", "", "Txt (*.txt)")

        if len(thresholdFile) != 1:
            return

        options = {     'reference' : self.segmentationRef,
                                'attributes': {'updateAtlas' : [None, None, thresholdFile[0]]},
                                'updateViewingObjects' : True,
                                'updateSettingWindow' : True}

        self.updateObject.emit(options)


    def addBundles(self):
        # Gets direction of bundle files
        bundleFiles, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select Bundle", "", "Bundles (*.bundles *.trk)")

        if len(bundleFiles) == 0:
            return

        options = {     'reference' : self.segmentationRef,
                                'attributes': {'updateAtlas' : [bundleFiles]},
                                'updateViewingObjects' : True,
                                'updateSettingWindow' : True}

        self.updateObject.emit(options)


    def removeBundles(self):
        pass


    @QtCore.pyqtSlot(int)
    def toggleCheck(self, row):
        tmpItem = self.ui.tableWidget.item(row,0)
        tmpItem.setCheckState(not tmpItem.checkState())


    # @QtCore.pyqtSlot()
    # def checkSegmentAll(self):
    #       self.ui.tableWidget.cellChanged.disconnect(self.toggleStates)

    #       visualize = self.ui.checkSegmentAll.isChecked()

    #       for i in range(self.ui.tableWidget.rowCount()):
    #               item = self.ui.tableWidget.item(i,0)
    #               item.setCheckState(visualize)
    #               self.states[self.bundleNames.index(item.text())] = visualize

    #       if self.ui.checkRender.isChecked():
    #               self.updateObject.emit(self.returnDict())

    #       self.ui.tableWidget.cellChanged.connect(self.toggleStates)


    # TMP!!!!!!
    @QtCore.pyqtSlot(str)
    def filter(self, text):
        self.ui.tableWidget.cellChanged.disconnect(self.toggleStates)

        self.ui.tableWidget.clearContents()
        self.ui.tableWidget.setRowCount(0)

        row = 0
        for i in range(len(self.bundleNames)):
            if self.bundleNames[i].find(text) != -1:
                self.ui.tableWidget.insertRow(row)
                item = QtWidgets.QTableWidgetItem(self.bundleNames[i])
                item.setCheckState(self.states[i])
                self.ui.tableWidget.setItem(row, 0, item)

                self.ui.tableWidget.setItem(row, 1, QtWidgets.QTableWidgetItem(self.bundleSize[i]))
                self.ui.tableWidget.item(row,1).setTextAlignment(QtCore.Qt.AlignRight)
                row = row + 1

        checkStateN = 0

        for i in range(self.ui.tableWidget.rowCount()):
            if self.ui.tableWidget.item(i,0).checkState():
                checkStateN += 1

        if checkStateN == self.ui.tableWidget.rowCount():
            self.ui.checkSegmentAll.setChecked(True)
        elif checkStateN == 0:
            self.ui.checkSegmentAll.setChecked(False)

        self.ui.tableWidget.cellChanged.connect(self.toggleStates)


    @QtCore.pyqtSlot(int)
    def toggleStates(self, row):
        item = self.ui.tableWidget.item(row,0)
        tmp = self.bundleNames.index(item.text())
        self.states[tmp] = bool(item.checkState())

        tmp = 0

        for i in range(self.ui.tableWidget.rowCount()):
            tmp += int(self.ui.tableWidget.item(i,0).checkState())

        if tmp == self.ui.tableWidget.rowCount():
            self.ui.checkSegmentAll.setChecked(True)
        else:
            self.ui.checkSegmentAll.setChecked(False)

        if self.ui.checkRender.isChecked():
            self.updateObject.emit(self.returnDict())


    @QtCore.pyqtSlot()
    def exportBundle(self):

        # Gets direction of bundle files
        outFile, fileType = QtWidgets.QFileDialog.getSaveFileName(self, "Save Bundle file", "", "Bundle (*.bundles)")#;;Trk file (*.trk)")

        options = {     'reference' : self.segmentationRef,
                                'attributes' : {'exportFile' : [outFile]}}

        self.updateObject.emit(options)


    @QtCore.pyqtSlot()
    def alphaChanged(self):
        alpha = self.alphaSlider.value()/(self.alphaSlider.maximum()+1)

        options = {     'reference' : self.segmentationRef,
                                'attributes': {'setAlpha' : [alpha]}}

        self.updateObject.emit(options)


    @QtCore.pyqtSlot()
    def applyMatrix(self):
        matrixFile, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select matrix", "", "matrix (*.npy *.trm)")

        if len(matrixFile) != 1:
            return

        options = {     'reference' : self.segmentationRef,
                                'attributes': {'updateAtlas' : [None, None, None, matrixFile[0]]},
                                'updateViewingObjects' : True,
                                'updateSettingWindow' : False}

        self.updateObject.emit(options)


    def updateWindow(self):

        validTableData = self.segmentationRef.atlas.getValidBundles()
        invalidTableData = self.segmentationRef.atlas.getInvalidBundles()

        self.ui.tableWidget.clearContents()
        self.ui.tableWidget.setRowCount(0)

        for i in range(len(validTableData)):
            self.ui.tableWidget.insertRow(i)
            data = validTableData[i]

            name = QtWidgets.QTableWidgetItem(data[0])
            nfiber = QtWidgets.QTableWidgetItem(str(data[1]))
            hie = QtWidgets.QTableWidgetItem()
            thr = QtWidgets.QTableWidgetItem(str(data[3]))

            hie.setBackground(QtGui.QBrush(QtGui.QColor(
                                                                                            int(data[2][0]*255),
                                                                                            int(data[2][1]*255),
                                                                                            int(data[2][2]*255))))

            self.ui.tableWidget.setItem(i, 0, name)
            self.ui.tableWidget.setItem(i, 1, nfiber)
            self.ui.tableWidget.setItem(i, 2, hie)
            self.ui.tableWidget.setItem(i, 3, thr)

        for i in range(len(invalidTableData)):
            j = i+len(validTableData)
            self.ui.tableWidget.insertRow(j)
            data = invalidTableData[i]

            name = QtWidgets.QTableWidgetItem(data[0])
            nfiber = QtWidgets.QTableWidgetItem(str(data[1]))
            hie = QtWidgets.QTableWidgetItem()
            thr = QtWidgets.QTableWidgetItem(str(data[3]))

            name.setBackground(QtGui.QBrush(QtGui.QColor(255,0,0)))
            nfiber.setBackground(QtGui.QBrush(QtGui.QColor(255,0,0)))
            hie.setBackground(QtGui.QBrush(QtGui.QColor(0,0,0)))
            thr.setBackground(QtGui.QBrush(QtGui.QColor(255,0,0)))

            self.ui.tableWidget.setItem(j, 0, name)
            self.ui.tableWidget.setItem(j, 1, nfiber)
            self.ui.tableWidget.setItem(j, 2, hie)
            self.ui.tableWidget.setItem(j, 3, thr)


    def returnDict(self):
        return {'reference' : self.segmentationRef,
                        'attributes' : {'setPercentage' : [self.ui.percentSlider.value()],
                                                        'setStates' : [self.states]},
                        'segmentSubject' : True}


    def __del__(self):
        print('Must clean befor leaving... INPLACESEGMENTATIONDIALOG', type(self))
        segmentationRef = None
        # super.__del__()



from enum import Enum


class VisualizationObject(Enum):
    NotDefined = -1
    Bundle = 0
    Mesh = 1
    Atlas = 7
    MRI = 2
    Segmentation = 3
    MRIVolume = 4
    MRISlice = 5
    ROI = 6


class ROIType(Enum):
    Sphere = 0
    Aabb = 1
    Obb = 2
    Plane = 3


class VisualizationActions(Enum):
    Delete = 0
    ToggleDrawable = 1
    Rotate = 2
    Translate = 3
    Scale = 4
    AddSegmentation = 5
    SliceModification = 6
    VolumeModification = 7
    MeshModification = 8
    ResetTransforms = 9
    LoadAndApplyMatrix = 10
    ShaderSelection = 11
    FocusObject = 12
    ColorSelection = 13


class SegmentationTypes(Enum):
    InPlace = 0
    ROIs = 1
    AtlasBased = 2
    FFClust = 3


segmentations = {'In place segmentation': SegmentationTypes.InPlace,
                 'Roi segmentation': SegmentationTypes.ROIs,
                 # 'Euclidean distance segmentation': SegmentationTypes.AtlasBased,
                 # 'FFClust': SegmentationTypes.FFClust
                 }

mriVisualizations = {'Volume render': VisualizationObject.MRIVolume,
                     'Slice': VisualizationObject.MRISlice}



'''
- Change tree from octree to R-tree. or study other balanced trees
- If octree pool overflows it should create a second pool
'''

import numpy as np
from .ROI import ROI
from .SegmentationHandler import SegmentationHandler
from ..BoundingBox import BoundingBox
from ..Tools.visualizationEnums import SegmentationTypes
from ..Tools.shunting_yard import shuntingYard
from ..Tools.DataStructures import OctreePointBased


class ROISegmentation(SegmentationHandler):
    # @averageTimeit
    def __init__(self, bundle, shaderDict):
        super().__init__(bundle, shaderDict)

        self.segmentationIdentifier = SegmentationTypes.ROIs

        self.tree = None
        self.roiValidator = []
        self.rois = []
        # self.fiberSizes = bundle.fiberSizes

        for sibling in self.parent.children:
            if isinstance(sibling, ROISegmentation):
                self.tree = sibling.tree

        if self.tree == None:
            self.tree = OctreePointBased(self.points, self.fiberSizes)

        self.fileName = 'ROI Segmentation' # temporal

        self.alpha = 0.8
        self.validLogic = False
        self.last_valid_logic = ''

        self.configFiberValidator()

        self._loadBuffers()
        self.buildVertex2Fiber()
        self.vboAndLinkVertex2Fiber()

        self.boundingbox = BoundingBox(shaderDict, self, bundle.boundingbox.dims, bundle.boundingbox.center)


    # @timeit
    def segmentMethod(self):
        rois2beQuery = None
        roisResults = None

        if self.validLogic:
            rois2beQuery = [self.rois[i] for i in self.logicRois]
        else:
            rois2beQuery = [self.rois[i] for i in [i for i, e in enumerate(self.roiValidator) if e]]


        n = len(rois2beQuery)

        if n == 0:
            self.fiberValidator[:self.curvescount] = 1

        else:
            roisResults = np.zeros((n, self.curvescount), dtype=np.int8)

            dt = np.dtype([     ('center', np.float32, 3),
                            ('radius', np.float32, 3),
                            ('roiType', np.int32, (1,))])

            dataPacked = np.empty(len(rois2beQuery), dtype=dt)

            for i in range(len(rois2beQuery)):
                dataPacked[i]['center'] = rois2beQuery[i].getCenter(self.inverseModel)
                dataPacked[i]['radius'] = rois2beQuery[i].getRadius(self.inverseModel)
                dataPacked[i]['roiType'] = rois2beQuery[i].getROIValue()


            self.tree.queryCollision(dataPacked, roisResults)


            if self.validLogic:
                infix = [roisResults[self.logicRois.index(i)] if isinstance(i, int) else i for i in self.logicInfix]
                self.fiberValidator[:self.curvescount] = shuntingYard(infix)

            else:
                self.fiberValidator[:self.curvescount] = roisResults.sum(axis=0, dtype=np.int8) #could overflow with 256 rois or more


    def addROI(self, roi):
        '''
        '''

        if isinstance(roi, ROI) and not roi in self.rois:
            self.children.append(roi)
            self.rois.append(roi)
            self.roiValidator.append(True)


    def removeROIFromIndex(self, index):
        roi = self.rois.pop(index)
        self.roiValidator.pop(index)
        self.children.remove(roi)


    def setValidatorAtIndex(self, index, validate):
        self.roiValidator[index] = validate


    def updateLogic(self, logicStr, reset_field=None):
        try:
            if logicStr == '':
                raise ValueError('ROISegmentation: Empty logic string.')
            if not self.rois:
                raise ValueError('Add at least one ROI.')
            parsed = self.parseLogicString(logicStr)
            if not parsed:
                raise ValueError('ROISegmentation: Not a valid string'
                                'for logic operations.\n'
                                'Valid operators are "| & ^ ( ) !".')
            self.logicInfix = parsed
            self.logicRois = list(set([i for i in parsed if isinstance(i, int)]))
            if max(self.logicRois) >= len(self.rois):
                raise ValueError(
                    'ROISegmentation: Not a valid ID for ROI selection. '
                    f'ID: {max(self.logicRois)}.\n'
                    f'Valid ID are in the interval [0, {len(self.rois) - 1}].')
            else:
                self.last_valid_logic = logicStr
                self.validLogic = True
        except ValueError:
            if callable(reset_field):
                reset_field(self.last_valid_logic)
            raise


    def parseLogicString(self, logicStr):
        ''' Could get a list of valid ids '''
        numberStack = ''
        infix = []

        for i in logicStr:
            if i.isdigit():
                numberStack += i

            elif i == '|' or i == '&' or i == '^' or i == '(' or i == ')' or i == '!':

                if numberStack != '':
                    infix.append(int(numberStack))
                    if infix[-1] >= len(self.rois):
                        return False
                    numberStack = ''

                infix.append(i)

            elif i == '+':
                if numberStack != '':
                    infix.append(int(numberStack))
                    if infix[-1] >= len(self.rois):
                        return False
                    numberStack = ''

                infix.append('|')

            else:
                return False

        if numberStack != '':
            infix.append(int(numberStack))

        return infix



from typing import Callable
from array import array


readTrkHeader: Callable[[str], dict]
readTrkBody: Callable[[str], dict]
readTrk: Callable[[str], dict]
readBundle: Callable[[str, int, array, int], dict]
readTck: Callable[[str], dict]
applyMatrix: Callable[[array, array], None]
reCalculateNormals: Callable[[array, array, array, int], None]
inPlaceSegmentationMethod: Callable[[int, int, array, array, array], None]
ROIsSegmentationPopulateAndDefragmentPool: Callable[[array, int, array, int, array, array, array], int]
ROIsSegmentationQueryOctree: Callable[[int, array, array, array, array, array, array, int], None]
ROISegmentationCreateEBO: Callable[[int, array, array, array, int], int]
ROISegmentationExportBundlesdata: Callable[[str, int, array, array, array, array, array], None]
AtlasBasedSegmentation: Callable[[array, int, array, int, int, array, array, array], None]
AtlasBasedSegmentationExportbundlesdata: Callable[[str, int, array, array, array, int, array], None]
reSampleBundle: Callable[[array, array, int, array, int], None]
ffclust: Callable[[int, array, array, float, int, int], array]

from .FiberVis_core import readTrkHeader,\
    readTrkBody,\
    readTrk,\
    readBundle,\
    readTck,\
    applyMatrix,\
    reCalculateNormals,\
    inPlaceSegmentationMethod,\
    ROIsSegmentationPopulateAndDefragmentPool,\
    ROIsSegmentationQueryOctree,\
    ROISegmentationCreateEBO,\
    ROISegmentationExportBundlesdata,\
    AtlasBasedSegmentation,\
    AtlasBasedSegmentationExportbundlesdata,\
    reSampleBundle,\
    ffclust



''' Base class for every drawable object
'''

from functools import wraps
from OpenGL import GL
import numpy as np
from .Tools import glm
from .Tools.visualizationEnums import VisualizationObject


fSize = np.dtype(np.float32).itemsize
iSize = np.dtype(np.int32).itemsize
uSize = np.dtype(np.uint32).itemsize


def propagateToChildren(func):
    """ Decorator for propagate the drawing function to the object's children.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        funcsName = func.__name__

        for child in args[0].children:
            getattr(child, funcsName)(*args[:-1], **kwargs)

        return func(*args, **kwargs)

    return wrapper

def drawable(func):
    """ Decorator for checking the drawable flag state from the object
    that calls the draw function from any VisualizationBaseObject inherited class.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        if args[0].drawable is True:
            return func(*args, **kwargs)
    return wrapper


def config(func):
    """ Decorator for draw function in any VisualizationBaseObject inherited class.
    Configures the using program and the desired vao
    and unlinks the vao when it's finished
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        obj = args[0]
        obj.shader[obj.selectedShader].glUseProgram()
        obj.loadUniforms()
        if type(obj.vao) is not np.ndarray:
            GL.glBindVertexArray(obj.vao)
        else:
            GL.glBindVertexArray(obj.vao[obj.selectedShader])
        func(*args, **kwargs)   # the draw function
        GL.glBindVertexArray(0)
    return wrapper


class VisualizationBaseObject:
    def __init__(self, parent=None, shaderDic=None):

        # Parent and children
        self.parent = parent
        self.children = []
        self.identifier = VisualizationObject.NotDefined

        # OpenGL buffer references
        self.vbo = None
        self.ebo = None
        self.vao = None

        self.texture = None

        # OpenGL shader program handler
        if isinstance(shaderDic, dict):
            self.shader = shaderDic[type(self)]
        else:
            self.shader = shaderDic
        self.nShader = 1

        # Matrixs
        self.model = glm.identity()

        self.inverseModel = glm.identity()

        self.rotationMat = glm.identity()
        self.translateMat = glm.identity()
        self.scaleMat = glm.identity()

        # Flags
        self.drawable = False
        self.clean = False
        self.selectedShader = 0

        self.boundingbox = None


    def cleanOpenGL(self):
        print('Must clean befor leaving... ', type(self))


    def __del__(self):
        if not self.clean:
            self.cleanOpenGL()


    @propagateToChildren
    def clearRef(self):
        self.parent = None
        self.children.clear()


    def draw(self):
        pass


    def setCamera(self, newCamera):
        pass

    def get_center(self):
        raise NotImplementedError

    def get_size(self):
        raise NotImplementedError

    def loadUniforms(self):

        GL.glUniformMatrix4fv(self.shader[self.selectedShader].glGetUniformLocation(
                "M"), 1, GL.GL_TRUE, self.model.getA())


    def remove(self, item):
        self.children.remove(item)


    def rotate(self, center, angle, axis):
        axis = (np.append(axis, 1) * self.rotationMat).getA()[0, :3]
        rotate = glm.translateMatrix(center) * glm.rotateMatrix(angle, axis) * glm.translateMatrix(-center)

        self.rotationMat *= rotate

        self.model = self.translateMat*self.rotationMat*self.scaleMat
        self.inverseModel = np.linalg.inv(self.model)

    def spin(self, angle, axis):
        self.rotate(self.get_center(), angle, axis)


    def translate(self, vec):
        self.translateMat = glm.translateMatrix(vec)

        self.model = self.translateMat*self.rotationMat*self.scaleMat
        self.inverseModel = np.linalg.inv(self.model)


    def stackTranslate(self, vec):
        newTranslateMat = glm.translateMatrix(vec)

        self.translateMat = newTranslateMat*self.translateMat

        self.model = self.translateMat*self.rotationMat*self.scaleMat
        self.inverseModel = np.linalg.inv(self.model)


    def scale(self, vec, center):
        self.scaleMat = glm.translateMatrix(center)*glm.scaleMatrix(vec)*glm.translateMatrix(-center)

        self.model = self.translateMat*self.rotationMat*self.scaleMat
        self.inverseModel = np.linalg.inv(self.model)


    def stackScale(self, vec, center):
        newScaleMat = glm.translateMatrix(center)*glm.scaleMatrix(vec)*glm.translateMatrix(-center)

        self.scaleMat = newScaleMat*self.scaleMat

        self.model = self.translateMat*self.rotationMat*self.scaleMat
        self.inverseModel = np.linalg.inv(self.model)


    def resetModel(self):
        self.scaleMat = glm.identity()

        self.translateMat = glm.identity()

        self.rotationMat = glm.identity()

        self.model = glm.identity()
        self.inverseModel = glm.identity()


    def setSelectedShader(self, newShaderId):
        if newShaderId < self.nShader:
            self.selectedShader = newShaderId
        else:
            raise TypeError('Unsupported shaderID, maximun is ', self.nShader, ' got ', newShaderId, '.')

    def setDrawBB(self, boolean):
        if self.boundingbox is not None:
            self.boundingbox.drawable = boolean
